"""
SPDX-License-Identifier: MIT

  Copyright (c) 2021, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import datetime
import json
import os
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional

import importlib_resources
from progress.bar import Bar
from progress.spinner import Spinner
from pypac.parser import PACFile

from scanoss.file_filters import FileFilters

from . import __version__
from .csvoutput import CsvOutput
from .cyclonedx import CycloneDx
from .scancodedeps import ScancodeDeps
from .scanoss_settings import ScanossSettings
from .scanossapi import ScanossApi
from .scanossbase import ScanossBase
from .scanossgrpc import ScanossGrpc
from .scanpostprocessor import ScanPostProcessor
from .scantype import ScanType
from .spdxlite import SpdxLite
from .threadeddependencies import SCOPE, ThreadedDependencies
from .threadedscanning import ThreadedScanning

FAST_WINNOWING = False
try:
    from scanoss_winnowing.winnowing import Winnowing

    FAST_WINNOWING = True
except (ModuleNotFoundError, ImportError):
    FAST_WINNOWING = False
    from .winnowing import Winnowing

WFP_FILE_START = 'file='
MAX_POST_SIZE = 64 * 1024  # 64k Max post size


class Scanner(ScanossBase):
    """
    SCANOSS scanning class
    Handle the scanning of files, snippets and dependencies
    """

    def __init__( # noqa: PLR0913, PLR0915
        self,
        wfp: str = None,
        scan_output: str = None,
        output_format: str = 'plain',
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
        api_key: str = None,
        url: str = None,
        flags: str = None,
        nb_threads: int = 5,
        post_size: int = 32,
        timeout: int = 180,
        no_wfp_file: bool = False,
        all_extensions: bool = False,
        all_folders: bool = False,
        hidden_files_folders: bool = False,
        scan_options: int = 7,
        sc_timeout: int = 600,
        sc_command: str = None,
        grpc_url: str = None,
        obfuscate: bool = False,
        ignore_cert_errors: bool = False,
        proxy: str = None,
        grpc_proxy: str = None,
        ca_cert: str = None,
        pac: PACFile = None,
        retry: int = 5,
        hpsm: bool = False,
        skip_size: int = 0,
        skip_extensions=None,
        skip_folders=None,
        strip_hpsm_ids=None,
        strip_snippet_ids=None,
        skip_md5_ids=None,
        scan_settings: 'ScanossSettings | None' = None,
        req_headers: dict = None,
    ):
        """
        Initialise scanning class, including Winnowing, ScanossApi, ThreadedScanning
        """
        super().__init__(debug, trace, quiet)
        if skip_folders is None:
            skip_folders = []
        if skip_extensions is None:
            skip_extensions = []
        self.wfp = wfp if wfp else 'scanner_output.wfp'
        self.scan_output = scan_output
        self.output_format = output_format
        self.no_wfp_file = no_wfp_file
        self.isatty = sys.stderr.isatty()
        self.all_extensions = all_extensions
        self.all_folders = all_folders
        self.hidden_files_folders = hidden_files_folders
        self.scan_options = scan_options
        self._skip_snippets = True if not scan_options & ScanType.SCAN_SNIPPETS.value else False
        self.hpsm = hpsm
        self.skip_folders = skip_folders
        self.skip_size = skip_size
        self.skip_extensions = skip_extensions
        self.req_headers = req_headers
        ver_details = Scanner.version_details()

        self.winnowing = Winnowing(
            debug=debug,
            quiet=quiet,
            skip_snippets=self._skip_snippets,
            all_extensions=all_extensions,
            obfuscate=obfuscate,
            hpsm=self.hpsm,
            strip_hpsm_ids=strip_hpsm_ids,
            strip_snippet_ids=strip_snippet_ids,
            skip_md5_ids=skip_md5_ids,
        )
        self.scanoss_api = ScanossApi(
            debug=debug,
            trace=trace,
            quiet=quiet,
            api_key=api_key,
            url=url,
            flags=flags,
            timeout=timeout,
            ver_details=ver_details,
            ignore_cert_errors=ignore_cert_errors,
            proxy=proxy,
            ca_cert=ca_cert,
            pac=pac,
            retry=retry,
            req_headers= self.req_headers,
        )
        sc_deps = ScancodeDeps(debug=debug, quiet=quiet, trace=trace, timeout=sc_timeout, sc_command=sc_command)
        grpc_api = ScanossGrpc(
            url=grpc_url,
            debug=debug,
            quiet=quiet,
            trace=trace,
            api_key=api_key,
            ver_details=ver_details,
            ca_cert=ca_cert,
            proxy=proxy,
            pac=pac,
            grpc_proxy=grpc_proxy,
            req_headers=self.req_headers,
        )
        self.threaded_deps = ThreadedDependencies(sc_deps, grpc_api, debug=debug, quiet=quiet, trace=trace)
        self.nb_threads = nb_threads
        if nb_threads and nb_threads > 0:
            self.threaded_scan = ThreadedScanning(
                self.scanoss_api, debug=debug, trace=trace, quiet=quiet, nb_threads=nb_threads
            )
        else:
            self.threaded_scan = None
        self.max_post_size = post_size * 1024 if post_size > 0 else MAX_POST_SIZE  # Set the max post size (default 64k)
        self.post_file_count = post_size if post_size > 0 else 32  # Max number of files for any given POST (default 32)
        if self._skip_snippets:
            self.max_post_size = 8 * 1024  # 8k Max post size if we're skipping snippets

        self.scan_settings = scan_settings
        self.post_processor = (
            ScanPostProcessor(scan_settings, debug=debug, trace=trace, quiet=quiet) if scan_settings else None
        )
        self._maybe_set_api_sbom()

    def _maybe_set_api_sbom(self):
        if not self.scan_settings:
            return
        sbom = self.scan_settings.get_sbom()
        if sbom:
            self.scanoss_api.set_sbom(sbom)

    @staticmethod
    def __count_files_in_wfp_file(wfp_file: str):
        """
        Count the number of files in the WFP that need to be processed
        Parameters
        ----------
            wfp_file: str
                WFP file to process
        """
        count = 0
        if wfp_file:
            with open(wfp_file) as f:
                for line in f:
                    if WFP_FILE_START in line:
                        count += 1
        return count

    @staticmethod
    def version_details() -> str:
        """
        Extract the date this version was produced
        :return: version creation date string
        """
        data = None
        try:
            f_name = importlib_resources.files(__name__) / 'data/build_date.txt'
            with importlib_resources.as_file(f_name) as f:
                with open(f, 'r', encoding='utf-8') as file:
                    data = file.read().rstrip()
        except Exception as e:
            Scanner.print_stderr(f'Warning: Problem loading build time details: {e}')
        if not data or len(data) == 0:
            now = datetime.datetime.now()
            data = f'date: {now.strftime("%Y%m%d%H%M%S")}, utime: {int(now.timestamp())}'
        return f'tool: scanoss-py, version: {__version__}, {data}'

    def __log_result(self, string, outfile=None):
        """
        Logs result to file or STDOUT
        """
        if not outfile and self.scan_output:
            outfile = self.scan_output
        if outfile:
            with open(outfile, 'a') as rf:
                rf.write(string + '\n')
        else:
            print(string)

    def is_file_or_snippet_scan(self):
        """
        Check if file or snippet scanning is enabled
        :return: True if enabled, False otherwise
        """
        if self.is_file_scan() or self.is_snippet_scan():
            return True
        return False

    def is_file_scan(self):
        """
        Check if file scanning is enabled
        :return: True if enabled, False otherwise
        """
        if self.scan_options & ScanType.SCAN_FILES.value:
            return True
        return False

    def is_snippet_scan(self):
        """
        Check if snippet scanning is enabled
        :return: True if enabled, False otherwise
        """
        if self.scan_options & ScanType.SCAN_SNIPPETS.value:
            return True
        return False

    def is_dependency_scan(self):
        """
        Check if dependency scanning is enabled
        :return: True if enabled, False otherwise
        """
        if self.scan_options & ScanType.SCAN_DEPENDENCIES.value:
            return True
        return False

    def scan_folder_with_options(  # noqa: PLR0913
        self,
        scan_dir: str,
        deps_file: str = None,
        file_map: dict = None,
        dep_scope: SCOPE = None,
        dep_scope_include: str = None,
        dep_scope_exclude: str = None,
    ) -> bool:
        """
        Scan the given folder for whatever scaning options that have been configured
        :param dep_scope_exclude: comma separated list of dependency scopes to exclude
        :param dep_scope_include: comma separated list of dependency scopes to include
        :param dep_scope: Enum dependency scope to use
        :param scan_dir: directory to scan
        :param deps_file: pre-parsed dependency file to decorate
        :param file_map: mapping of obfuscated files back into originals
        :return: True if successful, False otherwise
        """

        success = True
        if not scan_dir:
            raise Exception('ERROR: Please specify a folder to scan')
        if not os.path.exists(scan_dir) or not os.path.isdir(scan_dir):
            raise Exception(f'ERROR: Specified folder does not exist or is not a folder: {scan_dir}')
        if not self.is_file_or_snippet_scan() and not self.is_dependency_scan():
            raise Exception(f'ERROR: No scan options defined to scan folder: {scan_dir}')

        if self.scan_output:
            self.print_msg(f'Writing results to {self.scan_output}...')
        if self.is_dependency_scan():
            if not self.threaded_deps.run(
                what_to_scan=scan_dir,
                deps_file=deps_file,
                wait=False,
                dep_scope=dep_scope,
                dep_scope_include=dep_scope_include,
                dep_scope_exclude=dep_scope_exclude,
            ):  # Kick off a background dependency scan
                success = False
        if self.is_file_or_snippet_scan():
            if not self.scan_folder(scan_dir):
                success = False
        if self.threaded_scan:
            if not self.__finish_scan_threaded(file_map):
                success = False
        return success

    def scan_folder(self, scan_dir: str) -> bool:  # noqa: PLR0912, PLR0915
        """
        Scan the specified folder producing fingerprints, send to the SCANOSS API and return results

        :param scan_dir: str
                    Directory to scan
        :return True if successful, False otherwise
        """
        success = True
        if not scan_dir:
            raise Exception('ERROR: Please specify a folder to scan')
        if not os.path.exists(scan_dir) or not os.path.isdir(scan_dir):
            raise Exception(f'ERROR: Specified folder does not exist or is not a folder: {scan_dir}')

        file_filters = FileFilters(
            debug=self.debug,
            trace=self.trace,
            quiet=self.quiet,
            scanoss_settings=self.scan_settings,
            all_extensions=self.all_extensions,
            all_folders=self.all_folders,
            hidden_files_folders=self.hidden_files_folders,
            skip_size=self.skip_size,
            skip_folders=self.skip_folders,
            skip_extensions=self.skip_extensions,
            operation_type='scanning',
        )
        self.print_msg(f'Searching {scan_dir} for files to fingerprint...')
        spinner = None
        if not self.quiet and self.isatty:
            spinner = Spinner('Fingerprinting ')
        save_wfps_for_print = not self.no_wfp_file or not self.threaded_scan
        wfp_list = []
        scan_block = ''
        scan_size = 0
        queue_size = 0
        file_count = 0  # count all files fingerprinted
        wfp_file_count = 0  # count number of files in each queue post
        scan_started = False

        to_scan_files = file_filters.get_filtered_files_from_folder(scan_dir)
        for to_scan_file in to_scan_files:
            if self.threaded_scan and self.threaded_scan.stop_scanning():
                self.print_stderr('Warning: Aborting fingerprinting as the scanning service is not available.')
                break
            self.print_debug(f'Fingerprinting {to_scan_file}...')
            if spinner:
                spinner.next()
            abs_path = Path(scan_dir, to_scan_file).resolve()
            wfp = self.winnowing.wfp_for_file(str(abs_path), to_scan_file)
            if wfp is None or wfp == '':
                self.print_debug(f'No WFP returned for {to_scan_file}. Skipping.')
                continue
            if save_wfps_for_print:
                wfp_list.append(wfp)
            file_count += 1
            if self.threaded_scan:
                wfp_size = len(wfp.encode('utf-8'))
                # If the WFP is bigger than the max post size and we already have something stored in the scan block,
                # add it to the queue
                if scan_block != '' and (wfp_size + scan_size) >= self.max_post_size:
                    self.threaded_scan.queue_add(scan_block)
                    queue_size += 1
                    scan_block = ''
                    wfp_file_count = 0
                scan_block += wfp
                scan_size = len(scan_block.encode('utf-8'))
                wfp_file_count += 1
                # If the scan request block (group of WFPs) or larger than the POST size or we have reached the file limit, add it to the queue  # noqa: E501
                if wfp_file_count > self.post_file_count or scan_size >= self.max_post_size:
                    self.threaded_scan.queue_add(scan_block)
                    queue_size += 1
                    scan_block = ''
                    wfp_file_count = 0
                if not scan_started and queue_size > self.nb_threads:  # Start scanning if we have something to do
                    scan_started = True
                    if not self.threaded_scan.run(wait=False):
                        self.print_stderr('Warning: Some errors encounted while scanning. Results might be incomplete.')
                        success = False
        # End for loop
        if self.threaded_scan and scan_block != '':
            self.threaded_scan.queue_add(scan_block)  # Make sure all files have been submitted
        if spinner:
            spinner.finish()

        if file_count > 0:
            if save_wfps_for_print:  # Write a WFP file if no threading is requested
                self.print_debug(f'Writing fingerprints to {self.wfp}')
                with open(self.wfp, 'w') as f:
                    f.write(''.join(wfp_list))
            else:
                self.print_debug(f'Skipping writing WFP file {self.wfp}')
            if self.threaded_scan:
                success = self.__run_scan_threaded(scan_started, file_count)
        else:
            Scanner.print_stderr(f'Warning: No files found to scan in folder: {scan_dir}')
        return success

    def __run_scan_threaded(self, scan_started: bool, file_count: int) -> bool:
        """
        Start scanning the filtered files but do not wait for it to complete
        :param scan_started: If the scan has already started or not
        :param file_count:  Number of total files to be scanned
        :return: True if successful, False otherwise
        """
        success = True
        self.threaded_scan.update_bar(create=True, file_count=file_count)
        if not scan_started:
            if not self.threaded_scan.run(wait=False):  # Run the scan but do not wait for it to complete
                self.print_stderr('Warning: Some errors encounted while scanning. Results might be incomplete.')
                success = False
        return success

    def __finish_scan_threaded(self, file_map: Optional[Dict[Any, Any]] = None) -> bool:
        """Wait for the threaded scan to complete and process the results

        Args:
            file_map: Mapping of obfuscated files back to originals

        Returns:
            bool: True if successful, False otherwise

        Raises:
            ValueError: If output format is invalid
        """
        success: bool = True
        scan_responses = None
        dep_responses = None
        if self.is_file_or_snippet_scan():
            if not self.threaded_scan.complete():  # Wait for the scans to complete
                self.print_stderr('Warning: Scanning analysis ran into some trouble.')
                success = False
            self.threaded_scan.complete_bar()
            scan_responses = self.threaded_scan.responses
        if self.is_dependency_scan():
            self.print_msg('Retrieving dependency data...')
            if not self.threaded_deps.complete():
                self.print_stderr('Warning: Dependency analysis ran into some trouble.')
                success = False
            dep_responses = self.threaded_deps.responses

        raw_scan_results = self._merge_scan_results(scan_responses, dep_responses, file_map)

        if self.post_processor:
            results = self.post_processor.load_results(raw_scan_results).post_process()
        else:
            results = raw_scan_results

        if self.output_format == 'plain':
            self.__log_result(json.dumps(results, indent=2, sort_keys=True))
        elif self.output_format == 'cyclonedx':
            cdx = CycloneDx(self.debug, self.scan_output)
            success, _ = cdx.produce_from_json(results)
        elif self.output_format == 'spdxlite':
            spdxlite = SpdxLite(self.debug, self.scan_output)
            success = spdxlite.produce_from_json(results)
        elif self.output_format == 'csv':
            csvo = CsvOutput(self.debug, self.scan_output)
            success = csvo.produce_from_json(results)
        else:
            self.print_stderr(f'ERROR: Unknown output format: {self.output_format}')
            success = False
        return success

    def _merge_scan_results(
        self,
        scan_responses: Optional[List],
        dep_responses: Optional[Dict[str, Any]],
        file_map: Optional[Dict[str, Any]],
    ) -> Dict[str, Any]:
        """Merge scan and dependency responses into a single dictionary"""
        results: Dict[str, Any] = {}

        if scan_responses:
            for response in scan_responses:
                if response is not None:
                    if file_map:
                        response = self._deobfuscate_filenames(response, file_map)  # noqa: PLW2901
                    results.update(response)

        dep_files = dep_responses.get('files', None) if dep_responses else None
        if dep_files:
            for dep_file in dep_files:
                file = dep_file.pop('file', None)
                if file:
                    results[file] = [dep_file]

        return results

    def _deobfuscate_filenames(self, response: dict, file_map: dict) -> dict:
        """Convert obfuscated filenames back to original names"""
        deobfuscated = {}
        for key, value in response.items():
            deobfuscated_name = file_map.get(key, None)
            if deobfuscated_name:
                deobfuscated[deobfuscated_name] = value
            else:
                deobfuscated[key] = value
        return deobfuscated

    def scan_file_with_options(  # noqa: PLR0913
        self,
        file: str,
        deps_file: str = None,
        file_map: dict = None,
        dep_scope: SCOPE = None,
        dep_scope_include: str = None,
        dep_scope_exclude: str = None,
    ) -> bool:
        """
        Scan the given file for whatever scaning options that have been configured
        :param dep_scope:
        :param file: file to scan
        :param deps_file: pre-parsed dependency file to decorate
        :param file_map: mapping of obfuscated files back into originals
        :return: True if successful, False otherwise
        """
        success = True
        if not file:
            raise Exception('ERROR: Please specify a file to scan')
        if not os.path.exists(file) or not os.path.isfile(file):
            raise Exception(f'ERROR: Specified file does not exist or is not a file: {file}')
        if not self.is_file_or_snippet_scan() and not self.is_dependency_scan():
            raise Exception(f'ERROR: No scan options defined to scan file: {file}')

        if self.scan_output:
            self.print_msg(f'Writing results to {self.scan_output}...')
        if self.is_dependency_scan():
            if not self.threaded_deps.run(
                what_to_scan=file,
                deps_file=deps_file,
                wait=False,
                dep_scope=dep_scope,
                dep_scope_include=dep_scope_include,
                dep_scope_exclude=dep_scope_exclude,
            ):  # Kick off a background dependency scan
                success = False
        if self.is_file_or_snippet_scan():
            if not self.scan_file(file):
                success = False
        if self.threaded_scan:
            if not self.__finish_scan_threaded(file_map):
                success = False
        return success

    def scan_file(self, file: str) -> bool:
        """
        Scan the specified file and produce a result
        Parameters
        ----------
            file: str
                File to fingerprint and scan/identify
        :return True if successful, False otherwise
        """
        success = True
        if not file:
            raise Exception('ERROR: Please specify a file to scan')
        if not os.path.exists(file) or not os.path.isfile(file):
            raise Exception(f'ERROR: Specified files does not exist or is not a file: {file}')
        self.print_debug(f'Fingerprinting {file}...')
        wfp = self.winnowing.wfp_for_file(file, file)
        if wfp is not None and wfp != '':
            if self.threaded_scan:
                self.threaded_scan.queue_add(wfp)  # Submit the WFP for scanning
            self.print_debug(f'Scanning {file}...')
            if self.threaded_scan:
                success = self.__run_scan_threaded(False, 1)
        else:
            success = False
        return success

    def scan_files(self, files: []) -> bool:  # noqa: PLR0912, PLR0915
        """
        Scan the specified list of files, producing fingerprints, send to the SCANOSS API and return results
        Please note that by providing an explicit list you bypass any exclusions that may be defined on the scanner
        :param files: list[str]
                    List of filenames to scan
        :return True if successful, False otherwise
        """
        success = True
        if not files:
            raise Exception('ERROR: Please provide a non-empty list of filenames to scan')

        file_filters = FileFilters(
            debug=self.debug,
            trace=self.trace,
            quiet=self.quiet,
            scanoss_settings=self.scan_settings,
            all_extensions=self.all_extensions,
            all_folders=self.all_folders,
            hidden_files_folders=self.hidden_files_folders,
            skip_size=self.skip_size,
            skip_folders=self.skip_folders,
            skip_extensions=self.skip_extensions,
            operation_type='scanning',
        )
        spinner = None
        if not self.quiet and self.isatty:
            spinner = Spinner('Fingerprinting ')
        save_wfps_for_print = not self.no_wfp_file or not self.threaded_scan
        wfp_list = []
        scan_block = ''
        scan_size = 0
        queue_size = 0
        file_count = 0  # count all files fingerprinted
        wfp_file_count = 0  # count number of files in each queue post
        scan_started = False

        to_scan_files = file_filters.get_filtered_files_from_files(files)
        for file in to_scan_files:
            if self.threaded_scan and self.threaded_scan.stop_scanning():
                self.print_stderr('Warning: Aborting fingerprinting as the scanning service is not available.')
                break
            self.print_debug(f'Fingerprinting {file}...')
            if spinner:
                spinner.next()
            wfp = self.winnowing.wfp_for_file(file, file)
            if wfp is None or wfp == '':
                self.print_debug(f'No WFP returned for {file}. Skipping.')
                continue
            if save_wfps_for_print:
                wfp_list.append(wfp)
            file_count += 1
            if self.threaded_scan:
                wfp_size = len(wfp.encode('utf-8'))
                # If the WFP is bigger than the max post size and we already have something stored in the scan block, add it to the queue  # noqa: E501
                if scan_block != '' and (wfp_size + scan_size) >= self.max_post_size:
                    self.threaded_scan.queue_add(scan_block)
                    queue_size += 1
                    scan_block = ''
                    wfp_file_count = 0
                scan_block += wfp
                scan_size = len(scan_block.encode('utf-8'))
                wfp_file_count += 1
                # If the scan request block (group of WFPs) or larger than the POST size or we have reached the file limit, add it to the queue  # noqa: E501
                if wfp_file_count > self.post_file_count or scan_size >= self.max_post_size:
                    self.threaded_scan.queue_add(scan_block)
                    queue_size += 1
                    scan_block = ''
                    wfp_file_count = 0
                if not scan_started and queue_size > self.nb_threads:  # Start scanning if we have something to do
                    scan_started = True
                    if not self.threaded_scan.run(wait=False):
                        self.print_stderr(
                            'Warning: Some errors encounted while scanning. Results might be incomplete.'
                        )
                        success = False

        # End for loop
        if self.threaded_scan and scan_block != '':
            self.threaded_scan.queue_add(scan_block)  # Make sure all files have been submitted
        if spinner:
            spinner.finish()

        if file_count > 0:
            if save_wfps_for_print:  # Write a WFP file if no threading is requested
                self.print_debug(f'Writing fingerprints to {self.wfp}')
                with open(self.wfp, 'w') as f:
                    f.write(''.join(wfp_list))
            else:
                self.print_debug(f'Skipping writing WFP file {self.wfp}')
            if self.threaded_scan:
                success = self.__run_scan_threaded(scan_started, file_count)
        else:
            Scanner.print_stderr(f'Warning: No files found to scan from: {to_scan_files}')
        return success

    def scan_files_with_options(self, files: [], deps_file: str = None, file_map: dict = None) -> bool:
        """
        Scan the given list of files for whatever scaning options that have been configured
        :param files: list of files to scan
        :param deps_file: pre-parsed dependency file to decorate
        :param file_map: mapping of obfuscated files back into originals
        :return: True if successful, False otherwise
        """
        success = True
        if not files:
            raise Exception('ERROR: Please specify a list of files to scan')
        if not self.is_file_or_snippet_scan():
            raise Exception(f'ERROR: file or snippet scan options have to be set to scan files: {files}')
        if self.is_dependency_scan() or deps_file:
            raise Exception(
                'ERROR: The dependency scan option is currently not supported when scanning a list of files'
            )
        if self.scan_output:
            self.print_msg(f'Writing results to {self.scan_output}...')
        if self.is_file_or_snippet_scan():
            if not self.scan_files(files):
                success = False
        if self.threaded_scan:
            if not self.__finish_scan_threaded(file_map):
                success = False
        return success

    def scan_contents(self, filename: str, contents: bytes) -> bool:
        """
        Scan the given contents as a file

        :param filename: filename to associate with the contents
        :param contents: file contents
        :return: True if successful, False otherwise
        """
        success = True
        if not filename:
            raise Exception('ERROR: Please specify a filename to scan')
        if not contents:
            raise Exception('ERROR: Please specify a file contents to scan')

        self.print_debug(f'Fingerprinting {filename}...')
        wfp = self.winnowing.wfp_for_contents(filename, False, contents)
        if wfp is not None and wfp != '':
            if self.threaded_scan:
                self.threaded_scan.queue_add(wfp)  # Submit the WFP for scanning
            self.print_debug(f'Scanning {filename}...')
            if self.threaded_scan:
                success = self.__run_scan_threaded(False, 1)
        else:
            success = False
        if self.threaded_scan:
            if not self.__finish_scan_threaded():
                success = False
        return success

    def scan_wfp_file(self, file: str = None) -> bool:  # noqa: PLR0912, PLR0915
        """
        Scan the contents of the specified WFP file (in the current process)
        :param file: Scan the contents of the specified WFP file (in the current process)
        :return: True if successful, False otherwise
        """
        success = True
        wfp_file = file if file else self.wfp  # If a WFP file is specified, use it, otherwise us the default
        if not os.path.exists(wfp_file) or not os.path.isfile(wfp_file):
            raise Exception(f'ERROR: Specified WFP file does not exist or is not a file: {wfp_file}')
        file_count = Scanner.__count_files_in_wfp_file(wfp_file)
        cur_files = 0
        cur_size = 0
        batch_files = 0
        wfp = ''
        max_component = {'name': '', 'hits': 0}
        components = {}
        self.print_debug(f'Found {file_count} files to process.')
        raw_output = '{\n'
        file_print = ''
        bar = None
        if not self.quiet and self.isatty:
            bar = Bar('Scanning', max=file_count)
            bar.next(0)
        with open(wfp_file) as f:
            for line in f:
                if line.startswith(WFP_FILE_START):
                    if file_print:
                        wfp += file_print  # Store the WFP for the current file
                        cur_size = len(wfp.encode('utf-8'))
                    file_print = line  # Start storing the next file
                    cur_files += 1
                    batch_files += 1
                else:
                    file_print += line  # Store the rest of the WFP for this file
                l_size = cur_size + len(file_print.encode('utf-8'))
                # Hit the max post size, so sending the current batch and continue processing
                if l_size >= self.max_post_size and wfp:
                    self.print_debug(
                        f'Sending {batch_files} ({cur_files}) of'
                        f' {file_count} ({len(wfp.encode("utf-8"))} bytes) files to the ScanOSS API.'
                    )
                    if self.debug and cur_size > self.max_post_size:
                        Scanner.print_stderr(f'Warning: Post size {cur_size} greater than limit {self.max_post_size}')
                    scan_resp = self.scanoss_api.scan(wfp, max_component['name'])  # Scan current WFP and store
                    if bar:
                        bar.next(batch_files)
                    if scan_resp is not None:
                        for key, value in scan_resp.items():
                            raw_output += '  "%s":%s,' % (key, json.dumps(value, indent=2))
                            for v in value:
                                if hasattr(v, 'get'):
                                    if v.get('id') != 'none':
                                        vcv = '%s:%s:%s' % (v.get('vendor'), v.get('component'), v.get('version'))
                                        components[vcv] = components[vcv] + 1 if vcv in components else 1
                                        if max_component['hits'] < components[vcv]:
                                            max_component['name'] = v.get('component')
                                            max_component['hits'] = components[vcv]
                                else:
                                    Scanner.print_stderr(f'Warning: Unknown value: {v}')
                    else:
                        success = False
                    batch_files = 0
                    wfp = ''
        if file_print:
            wfp += file_print  # Store the WFP for the current file
        if wfp:
            self.print_debug(
                f'Sending {batch_files} ({cur_files}) of'
                f' {file_count} ({len(wfp.encode("utf-8"))} bytes) files to the ScanOSS API.'
            )
            scan_resp = self.scanoss_api.scan(wfp, max_component['name'])  # Scan current WFP and store
            if bar:
                bar.next(batch_files)
            first = True
            if scan_resp is not None:
                for key, value in scan_resp.items():
                    if first:
                        raw_output += '  "%s":%s' % (key, json.dumps(value, indent=2))
                        first = False
                    else:
                        raw_output += ',\n  "%s":%s' % (key, json.dumps(value, indent=2))
            else:
                success = False
        raw_output += '\n}'
        if bar:
            bar.finish()
        if self.output_format == 'plain':
            self.__log_result(raw_output)
        elif self.output_format == 'cyclonedx':
            cdx = CycloneDx(self.debug, self.scan_output)
            cdx.produce_from_str(raw_output)
        elif self.output_format == 'spdxlite':
            spdxlite = SpdxLite(self.debug, self.scan_output)
            success = spdxlite.produce_from_str(raw_output)
        elif self.output_format == 'csv':
            csvo = CsvOutput(self.debug, self.scan_output)
            csvo.produce_from_str(raw_output)
        else:
            self.print_stderr(f'ERROR: Unknown output format: {self.output_format}')
            success = False

        return success

    def scan_wfp_with_options(self, wfp: str, deps_file: str, file_map: dict = None) -> bool:
        """
        Scan the given WFP file for whatever scaning options that have been configured
        :param wfp: WFP file to scan
        :param deps_file: pre-parsed dependency file to decorate
        :param file_map: mapping of obfuscated files back into originals
        :return: True if successful, False otherwise
        """
        success = True
        wfp_file = wfp if wfp else self.wfp  # If a WFP file is specified, use it, otherwise us the default
        if not os.path.exists(wfp_file) or not os.path.isfile(wfp_file):
            raise Exception(f'ERROR: Specified WFP file does not exist or is not a file: {wfp_file}')

        if not self.is_file_or_snippet_scan() and not self.is_dependency_scan():
            raise Exception(f'ERROR: No scan options defined to scan WFP: {wfp}')

        if self.scan_output:
            self.print_msg(f'Writing results to {self.scan_output}...')
        if self.is_dependency_scan():
            if not self.threaded_deps.run(deps_file=deps_file, wait=False):  # Kick off a background dependency scan
                success = False
        if self.is_file_or_snippet_scan():
            if not self.scan_wfp_file_threaded(wfp_file):
                success = False
        if self.threaded_scan:
            if not self.__finish_scan_threaded(file_map):
                success = False
        return success

    def scan_wfp_file_threaded(self, file: str = None) -> bool:
        """
        Scan the contents of the specified WFP file (threaded)
        :param file: WFP file to scan (optional)
        return: True if successful, False otherwise
        """
        success = True
        wfp_file = file if file else self.wfp  # If a WFP file is specified, use it, otherwise us the default
        if not os.path.exists(wfp_file) or not os.path.isfile(wfp_file):
            raise Exception(f'ERROR: Specified WFP file does not exist or is not a file: {wfp_file}')
        cur_size = 0
        queue_size = 0
        file_count = 0  # count all files fingerprinted
        wfp_file_count = 0  # count number of files in each queue post
        scan_started = False
        wfp = ''
        scan_block = ''
        with open(wfp_file) as f:  # Parse the WFP file
            for line in f:
                if line.startswith(WFP_FILE_START):
                    if scan_block:
                        wfp += scan_block  # Store the WFP for the current file
                        cur_size = len(wfp.encode('utf-8'))
                    scan_block = line  # Start storing the next file
                    file_count += 1
                    wfp_file_count += 1
                else:
                    scan_block += line  # Store the rest of the WFP for this file
                l_size = cur_size + len(scan_block.encode('utf-8'))
                # Hit the max post size, so sending the current batch and continue processing
                if (wfp_file_count > self.post_file_count or l_size >= self.max_post_size) and wfp:
                    if self.debug and cur_size > self.max_post_size:
                        Scanner.print_stderr(f'Warning: Post size {cur_size} greater than limit {self.max_post_size}')
                    self.threaded_scan.queue_add(wfp)
                    queue_size += 1
                    wfp = ''
                    wfp_file_count = 0
                    if not scan_started and queue_size > self.nb_threads:  # Start scanning if we have something to do
                        scan_started = True
                        if not self.threaded_scan.run(wait=False):
                            self.print_stderr(
                                'Warning: Some errors uncounted while scanning. Results might be incomplete.'
                            )
                            success = False
            # End for loop
        if scan_block:
            wfp += scan_block  # Store the WFP for the current file
        if wfp:
            self.threaded_scan.queue_add(wfp)
            queue_size += 1

        if not self.__run_scan_threaded(scan_started, file_count):
            success = False
        return success

    def scan_wfp(self, wfp: str) -> bool:
        """
        Send the specified (single) WFP to ScanOSS for identification
        Parameters
        ----------
            wfp: str
                Winnowing Fingerprint to scan/identify
        """
        success = True
        if not wfp:
            raise Exception('ERROR: Please specify a WFP to scan')
        raw_output = '{\n'
        scan_resp = self.scanoss_api.scan(wfp)
        if scan_resp is not None:
            for key, value in scan_resp.items():
                raw_output += '  "%s":%s' % (key, json.dumps(value, indent=2))
        else:
            success = False
        raw_output += '\n}'
        if self.output_format == 'plain':
            self.__log_result(raw_output)
        elif self.output_format == 'cyclonedx':
            cdx = CycloneDx(self.debug, self.scan_output)
            cdx.produce_from_str(raw_output)
        elif self.output_format == 'spdxlite':
            spdxlite = SpdxLite(self.debug, self.scan_output)
            success = spdxlite.produce_from_str(raw_output)
        elif self.output_format == 'csv':
            csvo = CsvOutput(self.debug, self.scan_output)
            csvo.produce_from_str(raw_output)
        else:
            self.print_stderr(f'ERROR: Unknown output format: {self.output_format}')
            success = False

        return success

    def wfp_contents(self, filename: str, contents: bytes, wfp_file: str = None):
        """
        Fingerprint the specified contents as a file

        :param filename: filename to associate with the contents
        :param contents: file contents
        :param wfp_file: WFP to write results to (optional)
        :return:
        """
        if not filename:
            raise Exception('ERROR: Please specify a filename to scan')
        if not contents:
            raise Exception('ERROR: Please specify a file contents to scan')

        self.print_debug(f'Fingerprinting {filename}...')
        wfp = self.winnowing.wfp_for_contents(filename, False, contents)
        if wfp:
            if wfp_file:
                self.print_stderr(f'Writing fingerprints to {wfp_file}')
                with open(wfp_file, 'w') as f:
                    f.write(wfp)
            else:
                print(wfp)
        else:
            Scanner.print_stderr(f'Warning: No fingerprints generated for: {wfp_file}')

    def wfp_file(self, scan_file: str, wfp_file: str = None):
        """
        Fingerprint the specified file
        """
        if not scan_file:
            raise Exception('ERROR: Please specify a file to fingerprint')
        if not os.path.exists(scan_file) or not os.path.isfile(scan_file):
            raise Exception(f'ERROR: Specified file does not exist or is not a file: {scan_file}')

        self.print_debug(f'Fingerprinting {scan_file}...')
        wfp = self.winnowing.wfp_for_file(scan_file, scan_file)
        if wfp:
            if wfp_file:
                self.print_stderr(f'Writing fingerprints to {wfp_file}')
                with open(wfp_file, 'w') as f:
                    f.write(wfp)
            else:
                print(wfp)
        else:
            Scanner.print_stderr(f'Warning: No fingerprints generated for: {scan_file}')

    def wfp_folder(self, scan_dir: str, wfp_file: str = None):
        """
        Fingerprint the specified folder producing fingerprints
        """
        if not scan_dir:
            raise Exception('ERROR: Please specify a folder to fingerprint')
        if not os.path.exists(scan_dir) or not os.path.isdir(scan_dir):
            raise Exception(f'ERROR: Specified folder does not exist or is not a folder: {scan_dir}')
        file_filters = FileFilters(
            debug=self.debug,
            trace=self.trace,
            quiet=self.quiet,
            scanoss_settings=self.scan_settings,
            all_extensions=self.all_extensions,
            all_folders=self.all_folders,
            hidden_files_folders=self.hidden_files_folders,
            skip_size=self.skip_size,
            skip_folders=self.skip_folders,
            skip_extensions=self.skip_extensions,
            operation_type='scanning',
        )
        wfps = ''
        self.print_msg(f'Searching {scan_dir} for files to fingerprint...')
        spinner = None
        if not self.quiet and self.isatty:
            spinner = Spinner('Fingerprinting ')

        to_fingerprint_files = file_filters.get_filtered_files_from_folder(scan_dir)
        for file in to_fingerprint_files:
            if spinner:
                spinner.next()
            abs_path = Path(scan_dir, file).resolve()
            self.print_debug(f'Fingerprinting {file}...')
            wfps += self.winnowing.wfp_for_file(str(abs_path), file)
        if spinner:
            spinner.finish()
        if wfps:
            if wfp_file:
                self.print_stderr(f'Writing fingerprints to {wfp_file}')
                with open(wfp_file, 'w') as f:
                    f.write(wfps)
            else:
                print(wfps)
        else:
            Scanner.print_stderr(f'Warning: No files found to fingerprint in folder: {scan_dir}')


#
# End of ScanOSS Class
#



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: scanoss/api/common/v2/scanoss-common.proto
# Protobuf Python Version: 6.31.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    0,
    '',
    'scanoss/api/common/v2/scanoss-common.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from protoc_gen_openapiv2.options import annotations_pb2 as protoc__gen__openapiv2_dot_options_dot_annotations__pb2
from google.api import field_behavior_pb2 as google_dot_api_dot_field__behavior__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*scanoss/api/common/v2/scanoss-common.proto\x12\x15scanoss.api.common.v2\x1a.protoc-gen-openapiv2/options/annotations.proto\x1a\x1fgoogle/api/field_behavior.proto\"T\n\x0eStatusResponse\x12\x31\n\x06status\x18\x01 \x01(\x0e\x32!.scanoss.api.common.v2.StatusCode\x12\x0f\n\x07message\x18\x02 \x01(\t\"\x1e\n\x0b\x45\x63hoRequest\x12\x0f\n\x07message\x18\x01 \x01(\t\"\x1f\n\x0c\x45\x63hoResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"k\n\x10\x43omponentRequest\x12\x11\n\x04purl\x18\x01 \x01(\tB\x03\xe0\x41\x02\x12\x13\n\x0brequirement\x18\x02 \x01(\t:/\x92\x41,2*{\"purl\":\"pkg:github/scanoss/engine@1.0.0\"}\"\xc8\x01\n\x11\x43omponentsRequest\x12@\n\ncomponents\x18\x01 \x03(\x0b\x32\'.scanoss.api.common.v2.ComponentRequestB\x03\xe0\x41\x02:q\x92\x41n2l{\"components\":[{\"purl\":\"pkg:github/scanoss/engine@1.0.0\"},{\"purl\":\"pkg:github/scanoss/scanoss.py@v1.30.0\"}]}\"r\n\x0bPurlRequest\x12\x37\n\x05purls\x18\x01 \x03(\x0b\x32(.scanoss.api.common.v2.PurlRequest.Purls\x1a*\n\x05Purls\x12\x0c\n\x04purl\x18\x01 \x01(\t\x12\x13\n\x0brequirement\x18\x02 \x01(\t\")\n\x04Purl\x12\x0c\n\x04purl\x18\x01 \x01(\t\x12\x13\n\x0brequirement\x18\x02 \x01(\t*`\n\nStatusCode\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\x1b\n\x17SUCCEEDED_WITH_WARNINGS\x10\x02\x12\x0b\n\x07WARNING\x10\x03\x12\n\n\x06\x46\x41ILED\x10\x04\x42/Z-github.com/scanoss/papi/api/commonv2;commonv2b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'scanoss.api.common.v2.scanoss_common_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z-github.com/scanoss/papi/api/commonv2;commonv2'
  _globals['_COMPONENTREQUEST'].fields_by_name['purl']._loaded_options = None
  _globals['_COMPONENTREQUEST'].fields_by_name['purl']._serialized_options = b'\340A\002'
  _globals['_COMPONENTREQUEST']._loaded_options = None
  _globals['_COMPONENTREQUEST']._serialized_options = b'\222A,2*{\"purl\":\"pkg:github/scanoss/engine@1.0.0\"}'
  _globals['_COMPONENTSREQUEST'].fields_by_name['components']._loaded_options = None
  _globals['_COMPONENTSREQUEST'].fields_by_name['components']._serialized_options = b'\340A\002'
  _globals['_COMPONENTSREQUEST']._loaded_options = None
  _globals['_COMPONENTSREQUEST']._serialized_options = b'\222An2l{\"components\":[{\"purl\":\"pkg:github/scanoss/engine@1.0.0\"},{\"purl\":\"pkg:github/scanoss/scanoss.py@v1.30.0\"}]}'
  _globals['_STATUSCODE']._serialized_start=772
  _globals['_STATUSCODE']._serialized_end=868
  _globals['_STATUSRESPONSE']._serialized_start=150
  _globals['_STATUSRESPONSE']._serialized_end=234
  _globals['_ECHOREQUEST']._serialized_start=236
  _globals['_ECHOREQUEST']._serialized_end=266
  _globals['_ECHORESPONSE']._serialized_start=268
  _globals['_ECHORESPONSE']._serialized_end=299
  _globals['_COMPONENTREQUEST']._serialized_start=301
  _globals['_COMPONENTREQUEST']._serialized_end=408
  _globals['_COMPONENTSREQUEST']._serialized_start=411
  _globals['_COMPONENTSREQUEST']._serialized_end=611
  _globals['_PURLREQUEST']._serialized_start=613
  _globals['_PURLREQUEST']._serialized_end=727
  _globals['_PURLREQUEST_PURLS']._serialized_start=685
  _globals['_PURLREQUEST_PURLS']._serialized_end=727
  _globals['_PURL']._serialized_start=729
  _globals['_PURL']._serialized_end=770
# @@protoc_insertion_point(module_scope)



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2022, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import logging
import os
import sys
import time
from json.decoder import JSONDecodeError
import requests
import uuid
import http.client as http_client
import urllib3

from pypac import PACSession
from pypac.parser import PACFile
from urllib3.exceptions import InsecureRequestWarning

from .scanossbase import ScanossBase
from . import __version__


DEFAULT_URL = 'https://api.osskb.org/scan/direct'  # default free service URL
DEFAULT_URL2 = 'https://api.scanoss.com/scan/direct'  # default premium service URL
SCANOSS_SCAN_URL = os.environ.get('SCANOSS_SCAN_URL') if os.environ.get('SCANOSS_SCAN_URL') else DEFAULT_URL
SCANOSS_API_KEY = os.environ.get('SCANOSS_API_KEY') if os.environ.get('SCANOSS_API_KEY') else ''


class ScanossApi(ScanossBase):
    """
    ScanOSS REST API client class
    Currently support posting scan requests to the SCANOSS streaming API
    """

    def __init__(  # noqa: PLR0913, PLR0915
        self,
        scan_format: str = None,
        flags: str = None,
        url: str = None,
        api_key: str = None,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
        timeout: int = 180,
        ver_details: str = None,
        ignore_cert_errors: bool = False,
        proxy: str = None,
        ca_cert: str = None,
        pac: PACFile = None,
        retry: int = 5,
        req_headers: dict = None,
    ):
        """
        Initialise the SCANOSS API
        :param scan_format: Scan format (default plain)
        :param flags: Scanning flags (default None)
        :param url: API URL (default https://api.osskb.org/scan/direct)
        :param api_key: API Key (default None)
        :param debug: Enable debug (default False)
        :param trace: Enable trace (default False)
        :param quiet: Enable quite mode (default False)

        To set a custom certificate use:
            REQUESTS_CA_BUNDLE=/path/to/cert.pem
        To enable a Proxy use:
            HTTP_PROXY='http://<ip>:<port>'
            HTTPS_PROXY='http://<ip>:<port>'
        """
        super().__init__(debug, trace, quiet)
        self.url = url
        self.api_key = api_key
        self.sbom = None
        self.scan_format = scan_format if scan_format else 'plain'
        self.flags = flags
        self.timeout = timeout if timeout > 5 else 180
        self.retry_limit = retry if retry >= 0 else 5
        self.ignore_cert_errors = ignore_cert_errors
        self.req_headers = req_headers if req_headers else {}
        self.headers = {}

        if ver_details:
            self.headers['x-scanoss-client'] = ver_details
        if self.api_key:
            self.headers['X-Session'] = self.api_key
            self.headers['x-api-key'] = self.api_key
        self.headers['User-Agent'] = f'scanoss-py/{__version__}'
        self.headers['user-agent'] = f'scanoss-py/{__version__}'
        self.load_generic_headers()

        self.url = url if url else SCANOSS_SCAN_URL
        self.api_key = api_key if api_key else SCANOSS_API_KEY
        if self.api_key and not url and not os.environ.get('SCANOSS_SCAN_URL'):
            self.url = DEFAULT_URL2  # API key specific and no alternative URL, so use the default premium

        if self.trace:
            logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
            http_client.HTTPConnection.debuglevel = 1
        if pac and not proxy:  # Setup PAC session if requested (and no proxy has been explicitly set)
            self.print_debug('Setting up PAC session...')
            self.session = PACSession(pac=pac)
        else:
            self.session = requests.sessions.Session()
        self.verify = None
        if self.ignore_cert_errors:
            self.print_debug('Ignoring cert errors...')
            urllib3.disable_warnings(InsecureRequestWarning)
            self.verify = False
            self.session.verify = False
        elif ca_cert:
            self.verify = ca_cert
            self.session.verify = ca_cert
        self.proxies = {'https': proxy, 'http': proxy} if proxy else None
        if self.proxies:
            self.session.proxies = self.proxies

    def scan(self, wfp: str, context: str = None, scan_id: int = None):
        """
        Scan the specified WFP and return the JSON object
        :param wfp: WFP to scan
        :param context: Context to help with identification
        :param scan_id: ID of the scan being run (usually thread id)
        :return: JSON result object
        """
        request_id = str(uuid.uuid4())
        form_data = {}
        if self.sbom:
            form_data['type'] = self.sbom.get('scan_type')
            form_data['assets'] = self.sbom.get('assets')
        if self.scan_format:
            form_data['format'] = self.scan_format
        if self.flags:
            form_data['flags'] = self.flags
        if context:
            form_data['context'] = context

        scan_files = {'file': ('%s.wfp' % request_id, wfp)}
        headers = self.headers
        headers['x-request-id'] = request_id  # send a unique request id for each post
        r = None
        retry = 0  # Add some retry logic to cater for timeouts, etc.
        while retry <= self.retry_limit:
            retry += 1
            try:
                r = None
                r = self.session.post(
                    self.url, files=scan_files, data=form_data, headers=self.headers, timeout=self.timeout
                )
            except (requests.exceptions.SSLError, requests.exceptions.ProxyError) as e:
                self.print_stderr(f'ERROR: Exception ({e.__class__.__name__}) POSTing data - {e}.')
                raise Exception(f'ERROR: The SCANOSS API request failed for {self.url}') from e
            except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
                if retry > self.retry_limit:  # Timed out retry_limit or more times, fail
                    self.print_stderr(f'ERROR: {e.__class__.__name__} POSTing data ({request_id}) - {e}: {scan_files}')
                    raise Exception(
                        f'ERROR: The SCANOSS API request timed out ({e.__class__.__name__}) for {self.url}'
                    ) from e
                else:
                    self.print_stderr(f'Warning: {e.__class__.__name__} communicating with {self.url}. Retrying...')
                    time.sleep(5)
            except Exception as e:
                self.print_stderr(
                    f'ERROR: Exception ({e.__class__.__name__}) POSTing data ({request_id}) - {e}: {scan_files}'
                )
                raise Exception(f'ERROR: The SCANOSS API request failed for {self.url}') from e
            else:
                if r is None:
                    if retry > self.retry_limit:  # No response retry_limit or more times, fail
                        self.save_bad_req_wfp(scan_files, request_id, scan_id)
                        raise Exception(
                            f'ERROR: The SCANOSS API request ({request_id}) response object is empty for {self.url}'
                        )
                    else:
                        self.print_stderr(f'Warning: No response received from {self.url}. Retrying...')
                        time.sleep(5)
                elif r.status_code == 503:  # Service limits have most likely been reached
                    self.print_stderr(
                        f'ERROR: SCANOSS API rejected the scan request ({request_id}) due to '
                        f'service limits being exceeded'
                    )
                    self.print_stderr(f'ERROR: Details: {r.text.strip()}')
                    raise Exception(
                        f'ERROR: {r.status_code} - The SCANOSS API request ({request_id}) rejected '
                        f'for {self.url} due to service limits being exceeded.'
                    )
                elif r.status_code >= 400:
                    if retry > self.retry_limit:  # No response retry_limit or more times, fail
                        self.save_bad_req_wfp(scan_files, request_id, scan_id)
                        raise Exception(
                            f'ERROR: The SCANOSS API returned the following error: HTTP {r.status_code}, '
                            f'{r.text.strip()}'
                        )
                    else:
                        self.save_bad_req_wfp(scan_files, request_id, scan_id)
                        self.print_stderr(
                            f'Warning: Error response code {r.status_code} ({r.text.strip()}) from '
                            f'{self.url}. Retrying...'
                        )
                        time.sleep(5)
                else:
                    break  # Valid response, break out of the retry loop
        # End of while loop
        if r is None:
            self.save_bad_req_wfp(scan_files, request_id, scan_id)
            raise Exception(f'ERROR: The SCANOSS API request response object is empty for {self.url}')
        try:
            if 'xml' in self.scan_format:  # TODO remove XML parsing option?
                return r.text
            json_resp = r.json()
            return json_resp
        except (JSONDecodeError, Exception) as e:
            self.print_stderr(
                f'ERROR: The SCANOSS API returned an invalid JSON ({e.__class__.__name__} - {request_id}): {e}'
            )
            bad_json_file = f'bad_json-{scan_id}-{request_id}.txt' if scan_id else f'bad_json-{request_id}.txt'
            self.print_stderr(f'Ignoring result. Please look in "{bad_json_file}" for more details.')
            try:
                with open(bad_json_file, 'w') as f:
                    f.write(f'---Request ID Begin---\n{request_id}\n---Request ID End---\n')
                    f.write(f'---WFP Begin---\n{scan_files}\n---WFP End---\n---Bad JSON Begin---\n')
                    f.write(r.text)
                    f.write('---Bad JSON End---\n')
            except Exception as ee:
                self.print_stderr(
                    f'Warning: Issue writing bad json file - {bad_json_file} ({ee.__class__.__name__}): {ee}'
                )
            return None

    def save_bad_req_wfp(self, scan_files, request_id, scan_id):
        """
        Save the given WFP to a bad_request file
        :param scan_files: WFP
        :param request_id: request ID
        :param scan_id: scan thread id (optional)
        """
        bad_req_file = f'bad_request-{scan_id}-{request_id}.txt' if scan_id else f'bad_request-{request_id}.txt'
        try:
            self.print_stderr(
                f'No response object returned from API. Please look in "{bad_req_file}" for the offending WFP.'
            )
            with open(bad_req_file, 'w') as f:
                f.write(f'---Request ID Begin---\n{request_id}\n---Request ID End---\n')
                f.write(f'---WFP Begin---\n{scan_files}\n---WFP End---\n')
        except Exception as ee:
            self.print_stderr(
                f'Warning: Issue writing bad request file - {bad_req_file} ({ee.__class__.__name__}): {ee}'
            )

    def set_sbom(self, sbom):
        self.sbom = sbom
        return self

    def load_generic_headers(self):
        """
         Adds custom headers from req_headers to the headers collection.

         If x-api-key is present and no URL is configured (directly or via
         environment), sets URL to the premium endpoint (DEFAULT_URL2).
         """
        if self.req_headers:  # Load generic headers
            for key, value in self.req_headers.items():
                if key == 'x-api-key': # Set premium URL if x-api-key header is set
                    if not self.url and not os.environ.get('SCANOSS_SCAN_URL'):
                        self.url = DEFAULT_URL2  # API key specific and no alternative URL, so use the default premium
                    self.api_key = value
                self.headers[key] = value

#
# End of ScanossApi Class
#



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2021, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import json
import queue
import threading
from dataclasses import dataclass
from enum import Enum
from typing import Dict

from .scancodedeps import ScancodeDeps
from .scanossbase import ScanossBase
from .scanossgrpc import ScanossGrpc

DEP_FILE_PREFIX = 'file='  # Default prefix to signify an existing parsed dependency file

DEV_DEPENDENCIES = {
    'dev',
    'test',
    'development',
    'provided',
    'runtime',
    'devDependencies',
    'dev-dependencies',
    'testImplementation',
    'testCompile',
    'Test',
    'require-dev',
}


# Define an enum class
class SCOPE(Enum):
    PRODUCTION = 'prod'
    DEVELOPMENT = 'dev'


@dataclass
class ThreadedDependencies(ScanossBase):
    """ """

    inputs: queue.Queue = queue.Queue()
    output: queue.Queue = queue.Queue()

    def __init__(  # noqa: PLR0913
        self,
        sc_deps: ScancodeDeps,
        grpc_api: ScanossGrpc,
        what_to_scan: str = None,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
    ) -> None:
        """ """
        super().__init__(debug, trace, quiet)
        self.sc_deps = sc_deps
        self.grpc_api = grpc_api
        self.what_to_scan = what_to_scan
        self._thread = None
        self._errors = False

    @property
    def responses(self) -> Dict:
        """
        Get all responses back from the completed threads
        :return: JSON object
        """
        responses = list(self.output.queue)
        if responses:
            for resp in responses:
                return resp
        return None

    def run(
        self,
        what_to_scan: str = None,
        deps_file: str = None,
        wait: bool = True,
        dep_scope: SCOPE = None,
        dep_scope_include: str = None,
        dep_scope_exclude: str = None,
    ) -> bool:
        """
        Initiate a background scan for the specified file/dir
        :param dep_scope_exclude: comma separated list of dependency scopes to exclude
        :param dep_scope_include: comma separated list of dependency scopes to include
        :param dep_scope: Enum dependency scope to use
        :param what_to_scan: file/folder to scan
        :param deps_file: file to decorate instead of scan (overrides what_to_scan option)
        :param wait: wait for completion
        :return: True if successful, False if error encountered
        """
        what_to_scan = what_to_scan if what_to_scan else self.what_to_scan
        self._errors = False
        try:
            if deps_file:  # Decorate the given dependencies file
                self.print_msg(f'Decorating {deps_file} dependencies...')
                self.inputs.put(f'{DEP_FILE_PREFIX}{deps_file}')  # Add to queue and have parent wait on it
            else:  # Search for dependencies to decorate
                self.print_msg(f'Searching {what_to_scan} for dependencies...')
                self.inputs.put(what_to_scan)
                # Add to queue and have parent wait on it
            self._thread = threading.Thread(
                target=self.scan_dependencies(dep_scope, dep_scope_include, dep_scope_exclude), daemon=True
            )
            self._thread.start()
        except Exception as e:
            self.print_stderr(f'ERROR: Problem running threaded dependencies: {e}')
            self._errors = True
        if wait and not self._errors:  # Wait for all inputs to complete
            self.complete()
        return False if self._errors else True

    def filter_dependencies(self, deps, filter_dep) -> json:
        files = deps.get('files', [])
        # Iterate over files and their purls
        for file in files:
            if 'purls' in file:
                # Filter purls with scope 'dependencies' and remove the scope field
                file['purls'] = [
                    {key: value for key, value in purl.items() if key != 'scope'}
                    for purl in file['purls']
                    if filter_dep(purl.get('scope'))
                ]
        # End of for loop

        return {'files': [file for file in deps.get('files', []) if file.get('purls')]}

    def filter_dependencies_by_scopes(
        self, deps: json, dep_scope: SCOPE = None, dep_scope_include: str = None, dep_scope_exclude: str = None
    ) -> json:
        # Predefined set of scopes to filter

        # Include all scopes
        include_all = (dep_scope is None or dep_scope == '') and dep_scope_include is None and dep_scope_exclude is None
        ## All dependencies, remove scope key
        if include_all:
            return self.filter_dependencies(deps, lambda purl: True)

        # Use default list of scopes if a custom list is not set
        if (dep_scope is not None and dep_scope != '') and dep_scope_include is None and dep_scope_exclude is None:
            return self.filter_dependencies(
                deps,
                lambda purl: (dep_scope == SCOPE.PRODUCTION and purl not in DEV_DEPENDENCIES)
                or dep_scope == SCOPE.DEVELOPMENT
                and purl in DEV_DEPENDENCIES,
            )

        if (
            (dep_scope_include is not None and dep_scope_include != '')
            or dep_scope_exclude is not None
            and dep_scope_exclude != ''
        ):
            # Create sets from comma-separated strings, if provided
            exclude = set(dep_scope_exclude.split(',')) if dep_scope_exclude else set()
            include = set(dep_scope_include.split(',')) if dep_scope_include else set()

            # Define a lambda function that checks the inclusion/exclusion logic
            return self.filter_dependencies(
                deps, lambda purl: (exclude and purl not in exclude) or (not exclude and purl in include)
            )
        return None

    def scan_dependencies(  # noqa: PLR0912
        self, dep_scope: SCOPE = None, dep_scope_include: str = None, dep_scope_exclude: str = None
    ) -> None:
        """
        Scan for dependencies from the given file/dir or from an input file (from the input queue).
        """
        # TODO refactor to simplify branches based on PLR0912
        current_thread = threading.get_ident()
        self.print_trace(f'Starting dependency worker {current_thread}...')
        try:
            what_to_scan = self.inputs.get(timeout=5)  # Begin processing the dependency request
            deps = None
            if what_to_scan.startswith(DEP_FILE_PREFIX):  # We have a pre-parsed dependency file, load it
                deps = self.sc_deps.load_from_file(what_to_scan.strip(DEP_FILE_PREFIX))
            elif not self.sc_deps.run_scan(what_to_scan=what_to_scan):
                self._errors = True
            else:
                deps = self.sc_deps.produce_from_file()
                if dep_scope is not None:
                    self.print_debug(f'Filtering {dep_scope.name} dependencies')
                if dep_scope_include is not None:
                    self.print_debug(f"Including dependencies with '{dep_scope_include.split(',')}' scopes")
                if dep_scope_exclude is not None:
                    self.print_debug(f"Excluding dependencies with '{dep_scope_exclude.split(',')}' scopes")
                deps = self.filter_dependencies_by_scopes(deps, dep_scope, dep_scope_include, dep_scope_exclude)

            if not self._errors:
                if deps is None:
                    self.print_stderr(f'Problem searching for dependencies for: {what_to_scan}')
                    self._errors = True
                elif not deps or len(deps.get('files', [])) == 0:
                    self.print_debug(f'No dependencies found to decorate for: {what_to_scan}')
                else:
                    decorated_deps = self.grpc_api.get_dependencies(deps)
                    if decorated_deps:
                        self.output.put(decorated_deps)
                    else:
                        self._errors = True
        except Exception as e:
            self.print_stderr(f'ERROR: Problem encountered running dependency scan: {e}')
            self._errors = True
        finally:
            self.sc_deps.remove_interim_file()
            self.inputs.task_done()
        self.print_trace(f'Dependency thread complete ({current_thread}).')

    def complete(self) -> bool:
        """
        Wait for input queue to complete processing and complete the worker thread
        """
        try:
            self.inputs.join()
            self._thread.join(timeout=5)
        except Exception as e:
            self.print_stderr(f'WARNING: Issue encountered terminating dependency worker thread: {e}')
            self._errors = True
        return True if not self._errors else False


#
# End of ThreadedDependencies Class
#



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2022, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""
import csv
import json
import os.path
import sys

from .scanossbase import ScanossBase


class CsvOutput(ScanossBase):
    """
    CsvOutput management class
    Handle all interaction with CSV formatting
    """

    def __init__(self, debug: bool = False, output_file: str = None):
        """
        Initialise the CsvOutput class
        """
        super().__init__(debug)
        self.output_file = output_file
        self.debug = debug

    # TODO Refactor (fails linter)
    def parse(self, data: json): #noqa PLR0912, PLR0915
        """
        Parse the given input (raw/plain) JSON string and return CSV summary
        :param data: json - JSON object
        :return: CSV dictionary
        """
        if data is None:
            self.print_stderr('ERROR: No JSON data provided to parse.')
            return None
        if len(data) == 0:
            self.print_msg('Warning: Empty scan results provided. Returning empty CSV list.')
            return []
        self.print_debug('Processing raw results into CSV format...')
        csv_dict = []
        row_id = 1
        for f in data:
            file_details = data.get(f)
            # print(f'File: {f}: {file_details}')
            for d in file_details:
                id_details = d.get('id')
                if not id_details or id_details == 'none':
                    continue
                matched = d.get('matched', '')
                lines = d.get('lines', '').replace(',', ';')  # swap comma with semicolon to help basic parsers
                oss_lines = d.get('oss_lines', '').replace(',', ';')
                detected = {}
                if id_details == 'dependency':
                    dependencies = d.get('dependencies')
                    if not dependencies:
                        self.print_stderr(f'Warning: No Dependencies found for {f}: {file_details}')
                        continue
                    for deps in dependencies:
                        detected = {}
                        purl = deps.get('purl')
                        if not purl:
                            self.print_stderr(f'Warning: No PURL found for {f}: {deps}')
                            continue
                        detected['purls'] = purl
                        for field in ['component', 'version', 'latest', 'url']:
                            detected[field] = deps.get(field, '')
                        licenses = deps.get('licenses')
                        dc = []
                        if licenses:
                            for lic in licenses:
                                name = lic.get('name')
                                if name and name not in dc:  # Only save the license name once
                                    dc.append(name)
                        if not dc or len(dc) == 0:
                            detected['licenses'] = ''
                        else:
                            detected['licenses'] = ';'.join(dc)
                        # inventory_id,path,usage,detected_component,detected_license,
                        # detected_version,detected_latest,purl
                        csv_dict.append(
                            {
                                'inventory_id': row_id,
                                'path': f,
                                'detected_usage': id_details,
                                'detected_component': detected.get('component'),
                                'detected_license': detected.get('licenses'),
                                'detected_version': detected.get('version'),
                                'detected_latest': detected.get('latest'),
                                'detected_purls': detected.get('purls'),
                                'detected_url': detected.get('url'),
                                'detected_path': detected.get('file', ''),
                                'detected_match': matched,
                                'detected_lines': lines,
                                'detected_oss_lines': oss_lines,
                            }
                        )
                        row_id = row_id + 1
                else:
                    purls = d.get('purl')
                    if not purls:
                        self.print_stderr(f'Warning: Purl block missing for {f}: {file_details}')
                        continue
                    pa = []
                    for p in purls:
                        self.print_debug(f'Purl: {p}')
                        pa.append(p)
                    if not pa or len(pa) == 0:
                        self.print_stderr(f'Warning: No PURL found for {f}: {file_details}')
                        continue
                    detected['purls'] = ';'.join(pa)
                    for field in ['component', 'version', 'latest', 'url', 'file']:
                        detected[field] = d.get(field, '')
                    licenses = d.get('licenses')
                    dc = []
                    if licenses:
                        for lic in licenses:
                            name = lic.get('name')
                            if name and name not in dc:  # Only save the license name once
                                dc.append(lic.get('name'))
                    if not dc or len(dc) == 0:
                        detected['licenses'] = ''
                    else:
                        detected['licenses'] = ';'.join(dc)
                    # inventory_id,path,usage,detected_component,detected_license,detected_version,detected_latest,purl
                    csv_dict.append(
                        {
                            'inventory_id': row_id,
                            'path': f,
                            'detected_usage': id_details,
                            'detected_component': detected.get('component'),
                            'detected_license': detected.get('licenses'),
                            'detected_version': detected.get('version'),
                            'detected_latest': detected.get('latest'),
                            'detected_purls': detected.get('purls'),
                            'detected_url': detected.get('url'),
                            'detected_path': detected.get('file', ''),
                            'detected_match': matched,
                            'detected_lines': lines,
                            'detected_oss_lines': oss_lines,
                        }
                    )
                    row_id = row_id + 1
        return csv_dict

    def produce_from_file(self, json_file: str, output_file: str = None) -> bool:
        """
        Parse plain/raw input JSON file and produce CSV output
        :param json_file:
        :param output_file:
        :return: True if successful, False otherwise
        """
        if not json_file:
            self.print_stderr('ERROR: No JSON file provided to parse.')
            return False
        if not os.path.isfile(json_file):
            self.print_stderr(f'ERROR: JSON file does not exist or is not a file: {json_file}')
            return False
        with open(json_file, 'r') as f:
            success = self.produce_from_str(f.read(), output_file)
        return success

    def produce_from_json(self, data: json, output_file: str = None) -> bool:
        """
        Produce the CSV output from the input data
        :param data: JSON object
        :param output_file: Output file (optional)
        :return: True if successful, False otherwise
        """
        csv_data = self.parse(data)
        if csv_data is None:
            self.print_stderr('ERROR: No CSV data returned for the JSON string provided.')
            return False
        if len(csv_data) == 0:
            self.print_msg('Warning: Empty scan results - generating CSV with headers only.')
        # Header row/column details
        fields = [
            'inventory_id',
            'path',
            'detected_usage',
            'detected_component',
            'detected_license',
            'detected_version',
            'detected_latest',
            'detected_purls',
            'detected_url',
            'detected_match',
            'detected_lines',
            'detected_oss_lines',
            'detected_path',
        ]
        file = sys.stdout
        if not output_file and self.output_file:
            output_file = self.output_file
        if output_file:
            file = open(output_file, 'w')
        writer = csv.DictWriter(file, fieldnames=fields)
        writer.writeheader()  # writing headers (field names)
        writer.writerows(csv_data)  # writing data rows
        if output_file:
            file.close()

        return True

    def produce_from_str(self, json_str: str, output_file: str = None) -> bool:
        """
        Produce CSV output from input JSON string
        :param json_str: input JSON string
        :param output_file: Output file (optional)
        :return: True if successful, False otherwise
        """
        if not json_str:
            self.print_stderr('ERROR: No JSON string provided to parse.')
            return False
        try:
            data = json.loads(json_str)
        except Exception as e:
            self.print_stderr(f'ERROR: Problem parsing input JSON: {e}')
            return False
        return self.produce_from_json(data, output_file)


#
# End of CsvOutput Class
#



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: scanoss/api/licenses/v2/scanoss-licenses.proto
# Protobuf Python Version: 6.31.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    0,
    '',
    'scanoss/api/licenses/v2/scanoss-licenses.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from scanoss.api.common.v2 import scanoss_common_pb2 as scanoss_dot_api_dot_common_dot_v2_dot_scanoss__common__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from protoc_gen_openapiv2.options import annotations_pb2 as protoc__gen__openapiv2_dot_options_dot_annotations__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.scanoss/api/licenses/v2/scanoss-licenses.proto\x12\x17scanoss.api.licenses.v2\x1a*scanoss/api/common/v2/scanoss-common.proto\x1a\x1cgoogle/api/annotations.proto\x1a.protoc-gen-openapiv2/options/annotations.proto\"\xbd\x03\n\x18\x43omponentLicenseResponse\x12@\n\tcomponent\x18\x01 \x01(\x0b\x32-.scanoss.api.licenses.v2.ComponentLicenseInfo\x12\x35\n\x06status\x18\x02 \x01(\x0b\x32%.scanoss.api.common.v2.StatusResponse:\xa7\x02\x92\x41\xa3\x02\n\xa0\x02J\x9d\x02{\"component\":{\"purl\": \"pkg:github/scanoss/engine@1.0.0\", \"requirement\": \"\", \"version\": \"1.0.0\", \"statement\": \"GPL-2.0\", \"licenses\": [{\"id\": \"GPL-2.0\", \"full_name\": \"GNU General Public License v2.0 only\"}]}, \"status\": {\"status\": \"SUCCESS\", \"message\": \"Licenses Successfully retrieved\"}}\"\xe9\x04\n\x19\x43omponentsLicenseResponse\x12\x41\n\ncomponents\x18\x01 \x03(\x0b\x32-.scanoss.api.licenses.v2.ComponentLicenseInfo\x12\x35\n\x06status\x18\x02 \x01(\x0b\x32%.scanoss.api.common.v2.StatusResponse:\xd1\x03\x92\x41\xcd\x03\n\xca\x03J\xc7\x03{\"components\":[{\"purl\": \"pkg:github/scanoss/engine@1.0.0\", \"requirement\": \"\", \"version\": \"1.0.0\", \"statement\": \"GPL-2.0\", \"licenses\": [{\"id\": \"GPL-2.0\", \"full_name\": \"GNU General Public License v2.0 only\"}]}, {\"purl\": \"pkg:github/scanoss/scanoss.py@v1.30.0\",\"requirement\": \"\",\"version\": \"v1.30.0\",\"statement\": \"MIT\", \"licenses\": [{\"id\": \"MIT\",\"full_name\": \"MIT License\"}]}  ], \"status\": {\"status\": \"SUCCESS\", \"message\": \"Licenses Successfully retrieved\"}}\"\x89\x01\n\x16LicenseDetailsResponse\x12\x38\n\x07license\x18\x01 \x01(\x0b\x32\'.scanoss.api.licenses.v2.LicenseDetails\x12\x35\n\x06status\x18\x02 \x01(\x0b\x32%.scanoss.api.common.v2.StatusResponse\"\x81\x01\n\x13ObligationsResponse\x12\x33\n\x0bobligations\x18\x01 \x01(\x0b\x32\x1e.scanoss.api.licenses.v2.OSADL\x12\x35\n\x06status\x18\x02 \x01(\x0b\x32%.scanoss.api.common.v2.StatusResponse\"\xa3\x04\n\x04SPDX\x12\n\n\x02id\x18\x01 \x01(\t\x12\x11\n\tfull_name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65tails_url\x18\x04 \x01(\t\x12\x15\n\rreference_url\x18\x05 \x01(\t\x12\x15\n\ris_deprecated\x18\x06 \x01(\x08\x12\x14\n\x0cis_fsf_libre\x18\x07 \x01(\x08\x12\x17\n\x0fis_osi_approved\x18\x08 \x01(\x08\x12\x10\n\x08see_also\x18\t \x03(\t\x12>\n\ncross_refs\x18\n \x03(\x0b\x32*.scanoss.api.licenses.v2.SPDX.SPDXCrossRef\x12?\n\nexceptions\x18\x0b \x03(\x0b\x32+.scanoss.api.licenses.v2.SPDX.SPDXException\x1a\x88\x01\n\x0cSPDXCrossRef\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x10\n\x08is_valid\x18\x02 \x01(\x08\x12\x0f\n\x07is_live\x18\x03 \x01(\x08\x12\x11\n\ttimestamp\x18\x04 \x01(\t\x12\x17\n\x0fis_wayback_link\x18\x05 \x01(\x08\x12\r\n\x05order\x18\x06 \x01(\x05\x12\r\n\x05match\x18\x07 \x01(\t\x1al\n\rSPDXException\x12\n\n\x02id\x18\x01 \x01(\t\x12\x11\n\tfull_name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65tails_url\x18\x03 \x01(\t\x12\x10\n\x08see_also\x18\x05 \x03(\t\x12\x15\n\ris_deprecated\x18\x06 \x01(\x08\"\x97\x02\n\x05OSADL\x12\x17\n\x0f\x63opyleft_clause\x18\x01 \x01(\x08\x12\x14\n\x0cpatent_hints\x18\x02 \x01(\x08\x12\x15\n\rcompatibility\x18\x03 \x03(\t\x12\x1f\n\x17\x64\x65pending_compatibility\x18\x04 \x03(\t\x12\x17\n\x0fincompatibility\x18\x05 \x03(\t\x12>\n\tuse_cases\x18\x06 \x03(\x0b\x32+.scanoss.api.licenses.v2.OSADL.OSADLUseCase\x1aN\n\x0cOSADLUseCase\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x17\n\x0fobligation_text\x18\x02 \x01(\t\x12\x17\n\x0fobligation_json\x18\x03 \x01(\t\"{\n\x0bLicenseInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\x11\n\tfull_name\x18\x02 \x01(\t:M\x92\x41J\nHJF{\"id\": \"GPL-2.0\", \"full_name\": \"GNU General Public License v2.0 only\"}\"\xb3\x01\n\x0eLicenseDetails\x12\x11\n\tfull_name\x18\x01 \x01(\t\x12\x32\n\x04type\x18\x02 \x01(\x0e\x32$.scanoss.api.licenses.v2.LicenseType\x12+\n\x04spdx\x18\x03 \x01(\x0b\x32\x1d.scanoss.api.licenses.v2.SPDX\x12-\n\x05osadl\x18\x04 \x01(\x0b\x32\x1e.scanoss.api.licenses.v2.OSADL\"\x1c\n\x0eLicenseRequest\x12\n\n\x02id\x18\x01 \x01(\t\"\x95\x01\n\x14\x43omponentLicenseInfo\x12\x0c\n\x04purl\x18\x01 \x01(\t\x12\x13\n\x0brequirement\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x11\n\tstatement\x18\x04 \x01(\t\x12\x36\n\x08licenses\x18\x05 \x03(\x0b\x32$.scanoss.api.licenses.v2.LicenseInfo*l\n\x0bLicenseType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0e\n\nPERMISSIVE\x10\x01\x12\x0c\n\x08\x43OPYLEFT\x10\x02\x12\x0e\n\nCOMMERCIAL\x10\x03\x12\x0f\n\x0bPROPRIETARY\x10\x04\x12\x11\n\rPUBLIC_DOMAIN\x10\x05\x32\xd0\x05\n\x07License\x12q\n\x04\x45\x63ho\x12\".scanoss.api.common.v2.EchoRequest\x1a#.scanoss.api.common.v2.EchoResponse\" \x82\xd3\xe4\x93\x02\x1a\"\x15/api/v2/licenses/echo:\x01*\x12\x96\x01\n\x14GetComponentLicenses\x12\'.scanoss.api.common.v2.ComponentRequest\x1a\x31.scanoss.api.licenses.v2.ComponentLicenseResponse\"\"\x82\xd3\xe4\x93\x02\x1c\x12\x1a/api/v2/licenses/component\x12\x9d\x01\n\x15GetComponentsLicenses\x12(.scanoss.api.common.v2.ComponentsRequest\x1a\x32.scanoss.api.licenses.v2.ComponentsLicenseResponse\"&\x82\xd3\xe4\x93\x02 \"\x1b/api/v2/licenses/components:\x01*\x12\x88\x01\n\nGetDetails\x12\'.scanoss.api.licenses.v2.LicenseRequest\x1a/.scanoss.api.licenses.v2.LicenseDetailsResponse\" \x82\xd3\xe4\x93\x02\x1a\x12\x18/api/v2/licenses/details\x12\x8d\x01\n\x0eGetObligations\x12\'.scanoss.api.licenses.v2.LicenseRequest\x1a,.scanoss.api.licenses.v2.ObligationsResponse\"$\x82\xd3\xe4\x93\x02\x1e\x12\x1c/api/v2/licenses/obligationsB\xa7\x02Z1github.com/scanoss/papi/api/licensesv2;licensesv2\x92\x41\xf0\x01\x12\xb4\x01\n\x17SCANOSS License Service\x12\x46License service provides license intelligence for software components.\"L\n\x10scanoss-licenses\x12#https://github.com/scanoss/licenses\x1a\x13support@scanoss.com2\x03\x32.0\x1a\x0f\x61pi.scanoss.com*\x02\x01\x02\x32\x10\x61pplication/json:\x10\x61pplication/jsonb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'scanoss.api.licenses.v2.scanoss_licenses_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z1github.com/scanoss/papi/api/licensesv2;licensesv2\222A\360\001\022\264\001\n\027SCANOSS License Service\022FLicense service provides license intelligence for software components.\"L\n\020scanoss-licenses\022#https://github.com/scanoss/licenses\032\023support@scanoss.com2\0032.0\032\017api.scanoss.com*\002\001\0022\020application/json:\020application/json'
  _globals['_COMPONENTLICENSERESPONSE']._loaded_options = None
  _globals['_COMPONENTLICENSERESPONSE']._serialized_options = b'\222A\243\002\n\240\002J\235\002{\"component\":{\"purl\": \"pkg:github/scanoss/engine@1.0.0\", \"requirement\": \"\", \"version\": \"1.0.0\", \"statement\": \"GPL-2.0\", \"licenses\": [{\"id\": \"GPL-2.0\", \"full_name\": \"GNU General Public License v2.0 only\"}]}, \"status\": {\"status\": \"SUCCESS\", \"message\": \"Licenses Successfully retrieved\"}}'
  _globals['_COMPONENTSLICENSERESPONSE']._loaded_options = None
  _globals['_COMPONENTSLICENSERESPONSE']._serialized_options = b'\222A\315\003\n\312\003J\307\003{\"components\":[{\"purl\": \"pkg:github/scanoss/engine@1.0.0\", \"requirement\": \"\", \"version\": \"1.0.0\", \"statement\": \"GPL-2.0\", \"licenses\": [{\"id\": \"GPL-2.0\", \"full_name\": \"GNU General Public License v2.0 only\"}]}, {\"purl\": \"pkg:github/scanoss/scanoss.py@v1.30.0\",\"requirement\": \"\",\"version\": \"v1.30.0\",\"statement\": \"MIT\", \"licenses\": [{\"id\": \"MIT\",\"full_name\": \"MIT License\"}]}  ], \"status\": {\"status\": \"SUCCESS\", \"message\": \"Licenses Successfully retrieved\"}}'
  _globals['_LICENSEINFO']._loaded_options = None
  _globals['_LICENSEINFO']._serialized_options = b'\222AJ\nHJF{\"id\": \"GPL-2.0\", \"full_name\": \"GNU General Public License v2.0 only\"}'
  _globals['_LICENSE'].methods_by_name['Echo']._loaded_options = None
  _globals['_LICENSE'].methods_by_name['Echo']._serialized_options = b'\202\323\344\223\002\032\"\025/api/v2/licenses/echo:\001*'
  _globals['_LICENSE'].methods_by_name['GetComponentLicenses']._loaded_options = None
  _globals['_LICENSE'].methods_by_name['GetComponentLicenses']._serialized_options = b'\202\323\344\223\002\034\022\032/api/v2/licenses/component'
  _globals['_LICENSE'].methods_by_name['GetComponentsLicenses']._loaded_options = None
  _globals['_LICENSE'].methods_by_name['GetComponentsLicenses']._serialized_options = b'\202\323\344\223\002 \"\033/api/v2/licenses/components:\001*'
  _globals['_LICENSE'].methods_by_name['GetDetails']._loaded_options = None
  _globals['_LICENSE'].methods_by_name['GetDetails']._serialized_options = b'\202\323\344\223\002\032\022\030/api/v2/licenses/details'
  _globals['_LICENSE'].methods_by_name['GetObligations']._loaded_options = None
  _globals['_LICENSE'].methods_by_name['GetObligations']._serialized_options = b'\202\323\344\223\002\036\022\034/api/v2/licenses/obligations'
  _globals['_LICENSETYPE']._serialized_start=2858
  _globals['_LICENSETYPE']._serialized_end=2966
  _globals['_COMPONENTLICENSERESPONSE']._serialized_start=198
  _globals['_COMPONENTLICENSERESPONSE']._serialized_end=643
  _globals['_COMPONENTSLICENSERESPONSE']._serialized_start=646
  _globals['_COMPONENTSLICENSERESPONSE']._serialized_end=1263
  _globals['_LICENSEDETAILSRESPONSE']._serialized_start=1266
  _globals['_LICENSEDETAILSRESPONSE']._serialized_end=1403
  _globals['_OBLIGATIONSRESPONSE']._serialized_start=1406
  _globals['_OBLIGATIONSRESPONSE']._serialized_end=1535
  _globals['_SPDX']._serialized_start=1538
  _globals['_SPDX']._serialized_end=2085
  _globals['_SPDX_SPDXCROSSREF']._serialized_start=1839
  _globals['_SPDX_SPDXCROSSREF']._serialized_end=1975
  _globals['_SPDX_SPDXEXCEPTION']._serialized_start=1977
  _globals['_SPDX_SPDXEXCEPTION']._serialized_end=2085
  _globals['_OSADL']._serialized_start=2088
  _globals['_OSADL']._serialized_end=2367
  _globals['_OSADL_OSADLUSECASE']._serialized_start=2289
  _globals['_OSADL_OSADLUSECASE']._serialized_end=2367
  _globals['_LICENSEINFO']._serialized_start=2369
  _globals['_LICENSEINFO']._serialized_end=2492
  _globals['_LICENSEDETAILS']._serialized_start=2495
  _globals['_LICENSEDETAILS']._serialized_end=2674
  _globals['_LICENSEREQUEST']._serialized_start=2676
  _globals['_LICENSEREQUEST']._serialized_end=2704
  _globals['_COMPONENTLICENSEINFO']._serialized_start=2707
  _globals['_COMPONENTLICENSEINFO']._serialized_end=2856
  _globals['_LICENSE']._serialized_start=2969
  _globals['_LICENSE']._serialized_end=3689
# @@protoc_insertion_point(module_scope)



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2025, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import json

from .raw_base import RawBase


class ComponentSummary(RawBase):
    def _get_component_summary_from_components(self, scan_components: list)-> dict:
        """
        Get a component summary from detected components.

        :param components: List of all components
        :return: Dict with license summary information
        """
        # A component is considered unique by its combination of PURL (Package URL) and license
        component_licenses = self._group_components_by_license(scan_components)
        total_components = len(component_licenses)
        # Get undeclared components
        undeclared_components = len([c for c in component_licenses if c['status'] == 'pending'])

        components: list = []
        total_undeclared_files = 0
        total_files_detected = 0
        for component in scan_components:
            total_files_detected += component['count']
            total_undeclared_files += component['undeclared']
            components.append({
                'purl': component['purl'],
                'version': component['version'],
                'count': component['count'],
                'undeclared': component['undeclared'],
                'declared': component['count'] - component['undeclared'],
            })
        ## End for loop components
        return {
            "components": component_licenses,
            'totalComponents': total_components,
            'undeclaredComponents': undeclared_components,
            'declaredComponents': total_components - undeclared_components,
            'totalFilesDetected': total_files_detected,
            'totalFilesUndeclared': total_undeclared_files,
            'totalFilesDeclared': total_files_detected - total_undeclared_files,
        }

    def _get_components(self):
        """
        Extract and process components from results and their dependencies.

        This method performs the following steps:
        1. Validates that `self.results` is loaded. Returns `None` if not.
        2. Extracts file, snippet, and dependency components into a dictionary.
        3. Converts components to a list and processes their licenses.

        :return: A list of processed components with license data, or `None` if `self.results` is not set.
        """
        if self.results is None:
            raise ValueError(f'Error: No results found in ${self.filepath}')

        components: dict = {}
        # Extract component and license data from file and dependency results. Both helpers mutate `components`
        self._get_components_data(self.results, components)
        return self._convert_components_to_list(components)

    def run(self):
        components = self._get_components()
        component_summary = self._get_component_summary_from_components(components)
        self.print_to_file_or_stdout(json.dumps(component_summary, indent=2), self.output)
        return component_summary
#
# End of ComponentSummary Class
#


"""
SPDX-License-Identifier: MIT

  Copyright (c) 2022, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import csv
import os
import pathlib
import sys

from progress.spinner import Spinner

from .scanossbase import ScanossBase


class FileCount(ScanossBase):
    """
    SCANOSS File Type Count class
    Handle the scanning of files, snippets and dependencies
    """

    def __init__(
        self,
        scan_output: str = None,
        hidden_files_folders: bool = False,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
    ):
        """
        Initialise scanning class
        """
        super().__init__(debug, trace, quiet)
        self.scan_output = scan_output
        self.isatty = sys.stderr.isatty()
        self.hidden_files_folders = hidden_files_folders

    def __filter_files(self, files: list) -> list:
        """
        Filter which files should be considered for processing
        :param files: list of files to filter
        :return list of filtered files
        """
        file_list = []
        for f in files:
            ignore = False
            if f.startswith('.') and not self.hidden_files_folders:  # Ignore all . files unless requested
                ignore = True
            if not ignore:
                file_list.append(f)
        return file_list

    def __filter_dirs(self, dirs: list) -> list:
        """
        Filter which folders should be considered for processing
        :param dirs: list of directories to filter
        :return: list of filtered directories
        """
        dir_list = []
        for d in dirs:
            ignore = False
            if d.startswith('.') and not self.hidden_files_folders:  # Ignore all . folders unless requested
                ignore = True
            if not ignore:
                dir_list.append(d)
        return dir_list

    def __log_result(self, string, outfile=None):
        """
        Logs result to file or STDOUT
        """
        if not outfile and self.scan_output:
            outfile = self.scan_output
        if outfile:
            with open(outfile, 'a') as rf:
                rf.write(string + '\n')
        else:
            print(string)

    def count_files(self, scan_dir: str) -> bool:
        """
        Search the specified folder producing counting the file types found
        :param scan_dir str
                    Directory to scan
        :return True if successful, False otherwise
        """
        success = True
        if not scan_dir:
            raise Exception(f'ERROR: Please specify a folder to scan')
        if not os.path.exists(scan_dir) or not os.path.isdir(scan_dir):
            raise Exception(f'ERROR: Specified folder does not exist or is not a folder: {scan_dir}')

        self.print_msg(f'Searching {scan_dir} for files to count...')
        spinner = None
        if not self.quiet and self.isatty:
            spinner = Spinner('Searching ')
        file_types = {}
        file_count = 0
        file_size = 0
        for root, dirs, files in os.walk(scan_dir):
            self.print_trace(f'U Root: {root}, Dirs: {dirs}, Files {files}')
            dirs[:] = self.__filter_dirs(dirs)  # Strip out unwanted directories
            filtered_files = self.__filter_files(files)  # Strip out unwanted files
            self.print_trace(f'F Root: {root}, Dirs: {dirs}, Files {filtered_files}')
            for file in filtered_files:  # Cycle through each filtered file
                path = os.path.join(root, file)
                f_size = 0
                try:
                    f_size = os.stat(path).st_size
                except Exception as e:
                    self.print_trace(f'Ignoring missing symlink file: {file} ({e})')  # broken symlink
                if f_size > 0:  # Ignore broken links and empty files
                    file_count = file_count + 1
                    file_size = file_size + f_size
                    f_suffix = pathlib.Path(file).suffix
                    if not f_suffix or f_suffix == '':
                        f_suffix = 'no_suffix'
                    self.print_trace(f'Counting {path} ({f_suffix} - {f_size})..')
                    fc = file_types.get(f_suffix)
                    if not fc:
                        fc = [1, f_size]
                    else:
                        fc[0] = fc[0] + 1
                        fc[1] = fc[1] + f_size
                    file_types[f_suffix] = fc
                    if spinner:
                        spinner.next()
        # End for loop
        if spinner:
            spinner.finish()
        self.print_stderr(f'Found {file_count:,.0f} files with a total size of {file_size / (1 << 20):,.2f} MB.')
        if file_types:
            csv_dict = []
            for k in file_types:
                d = file_types[k]
                csv_dict.append({'extension': k, 'count': d[0], 'size(MB)': f'{d[1] / (1 << 20):,.2f}'})
            fields = ['extension', 'count', 'size(MB)']
            file = sys.stdout
            if self.scan_output:
                file = open(self.scan_output, 'w')
            writer = csv.DictWriter(file, fieldnames=fields)
            writer.writeheader()  # writing headers (field names)
            writer.writerows(csv_dict)  # writing data rows
            if self.scan_output:
                file.close()
        else:
            FileCount.print_stderr(f'Warning: No files found to count in folder: {scan_dir}')
        return success


#
# End of ScanOSS Class
#



import json
from dataclasses import dataclass
from typing import Dict, List, Optional

from scanoss.scanossbase import ScanossBase
from scanoss.scanossgrpc import ScanossGrpc
from scanoss.utils.abstract_presenter import AbstractPresenter
from scanoss.utils.file import validate_json_file


class ScanossCryptographyError(Exception):
    pass


MIN_SPLIT_PARTS = 2


@dataclass
class CryptographyConfig:
    purl: List[str]
    input_file: Optional[str] = None
    output_file: Optional[str] = None
    header: Optional[str] = None
    debug: bool = False
    trace: bool = False
    quiet: bool = False
    with_range: bool = False

    def __post_init__(self):
        """
        Validate that the configuration is valid.
        """
        if self.purl:
            if self.with_range:
                for purl in self.purl:
                    parts = purl.split('@')
                    if not (len(parts) >= MIN_SPLIT_PARTS and parts[1]):
                        raise ScanossCryptographyError(
                            f'Invalid PURL format: "{purl}".' f'It must include a version (e.g., pkg:type/name@version)'
                        )
        if self.input_file:
            input_file_validation = validate_json_file(self.input_file)
            if not input_file_validation.is_valid:
                raise ScanossCryptographyError(
                    f'There was a problem with the purl input file. {input_file_validation.error}'
                )
            if (
                not isinstance(input_file_validation.data, dict)
                or 'purls' not in input_file_validation.data
                or not isinstance(input_file_validation.data['purls'], list)
                or not all(isinstance(p, dict) and 'purl' in p for p in input_file_validation.data['purls'])
            ):
                raise ScanossCryptographyError('The supplied input file is not in the correct PurlRequest format.')
            purls = input_file_validation.data['purls']
            purls_with_requirement = []
            if self.with_range and any('requirement' not in p for p in purls):
                raise ScanossCryptographyError(
                    f'One or more PURLs in "{self.input_file}" are missing the "requirement" field.'
                )

            for purl in purls:
                if 'requirement' in purl:
                    purls_with_requirement.append(f'{purl["purl"]}@{purl["requirement"]}')
                else:
                    purls_with_requirement.append(purl['purl'])
            self.purl = purls_with_requirement


def create_cryptography_config_from_args(args) -> CryptographyConfig:
    return CryptographyConfig(
        debug=getattr(args, 'debug', False),
        trace=getattr(args, 'trace', False),
        quiet=getattr(args, 'quiet', False),
        with_range=getattr(args, 'with_range', False),
        purl=getattr(args, 'purl', []),
        input_file=getattr(args, 'input', None),
        output_file=getattr(args, 'output', None),
        header=getattr(args, 'header', None),
    )


class Cryptography:
    """
    Cryptography Class

    This class is used to decorate purls with cryptography information.
    """

    def __init__(
        self,
        config: CryptographyConfig,
        client: ScanossGrpc,
    ):
        """
        Initialize the Cryptography.

        Args:
            config (CryptographyConfig): Configuration parameters for the cryptography.
            client (ScanossGrpc): gRPC client for communicating with the scanning service.
        """
        self.base = ScanossBase(
            debug=config.debug,
            trace=config.trace,
            quiet=config.quiet,
        )
        self.presenter = CryptographyPresenter(
            self,
            debug=config.debug,
            trace=config.trace,
            quiet=config.quiet,
        )

        self.client = client
        self.config = config
        self.purls_request = self._build_purls_request()
        self.results = None

    def get_algorithms(self) -> Optional[Dict]:
        """
        Get the cryptographic algorithms for the provided purl or input file.

        Returns:
            Optional[Dict]: The folder hash response from the gRPC client, or None if an error occurs.
        """

        if not self.purls_request:
            raise ScanossCryptographyError('No PURLs supplied. Provide --purl or --input.')
        self.base.print_stderr(
            f'Getting cryptographic algorithms for {", ".join([p["purl"] for p in self.purls_request["purls"]])}'
        )
        if self.config.with_range:
            response = self.client.get_crypto_algorithms_in_range_for_purl(self.purls_request)
        else:
            response = self.client.get_crypto_algorithms_for_purl(self.purls_request)
        if response:
            self.results = response

        return self.results

    def get_encryption_hints(self) -> Optional[Dict]:
        """
        Get the encryption hints for the provided purl or input file.

        Returns:
            Optional[Dict]: The encryption hints response from the gRPC client, or None if an error occurs.
        """

        if not self.purls_request:
            raise ScanossCryptographyError('No PURLs supplied. Provide --purl or --input.')
        self.base.print_stderr(
            f'Getting encryption hints '
            f'{"in range" if self.config.with_range else ""} '
            f'for {", ".join([p["purl"] for p in self.purls_request["purls"]])}'
        )
        if self.config.with_range:
            response = self.client.get_encryption_hints_in_range_for_purl(self.purls_request)
        else:
            response = self.client.get_encryption_hints_for_purl(self.purls_request)
        if response:
            self.results = response

        return self.results

    def get_versions_in_range(self) -> Optional[Dict]:
        """
        Given a list of PURLS and version ranges, get a list of versions that do/do not contain cryptographic algorithms

        Returns:
            Optional[Dict]: The versions in range response from the gRPC client, or None if an error occurs.
        """

        if not self.purls_request:
            raise ScanossCryptographyError('No PURLs supplied. Provide --purl or --input.')

        self.base.print_stderr(
            f'Getting versions in range for {", ".join([p["purl"] for p in self.purls_request["purls"]])}'
        )

        response = self.client.get_versions_in_range_for_purl(self.purls_request)
        if response:
            self.results = response

        return self.results

    def _build_purls_request(
        self,
    ) -> Optional[dict]:
        """
        Load the specified purls from a JSON file or a list of PURLs and return a dictionary

        Args:
            json_file (Optional[str], optional): The JSON file containing the PURLs. Defaults to None.
            purls (Optional[List[str]], optional): The list of PURLs. Defaults to None.

        Returns:
            Optional[dict]: The dictionary containing the PURLs
        """
        return {
            'purls': [
                {
                    'purl': self._remove_version_from_purl(purl),
                    'requirement': self._extract_version_from_purl(purl),
                }
                for purl in self.config.purl
            ]
        }

    def _remove_version_from_purl(self, purl: str) -> str:
        """
        Remove version from purl

        Args:
            purl (str): The purl string to remove the version from

        Returns:
            str: The purl string without the version
        """
        if '@' not in purl:
            return purl
        return purl.split('@')[0]

    def _extract_version_from_purl(self, purl: str) -> str:
        """
        Extract version from purl

        Args:
            purl (str): The purl string to extract the version from

        Returns:
            str: The extracted version

        Raises:
            ScanossCryptographyError: If the purl is not in the correct format
        """
        try:
            return purl.split('@')[-1]
        except IndexError:
            raise ScanossCryptographyError(f'Invalid purl format: {purl}')

    def present(
        self,
        output_format: Optional[str] = None,
        output_file: Optional[str] = None,
    ):
        """Present the results in the selected format"""
        self.presenter.present(output_format=output_format, output_file=output_file)


class CryptographyPresenter(AbstractPresenter):
    """
    Cryptography presenter class
    Handles the presentation of the cryptography results
    """

    def __init__(self, cryptography: Cryptography, **kwargs):
        super().__init__(**kwargs)
        self.cryptography = cryptography

    def _format_json_output(self) -> str:
        """
        Format the scan output data into a JSON object

        Returns:
            str: The formatted JSON string
        """
        return json.dumps(self.cryptography.results, indent=2)

    def _format_plain_output(self) -> str:
        """
        Format the scan output data into a plain text string
        """
        return (
            json.dumps(self.cryptography.results, indent=2)
            if isinstance(self.cryptography.results, dict)
            else str(self.cryptography.results)
        )

    def _format_cyclonedx_output(self) -> str:
        raise NotImplementedError('CycloneDX output is not implemented')

    def _format_spdxlite_output(self) -> str:
        raise NotImplementedError('SPDXlite output is not implemented')

    def _format_csv_output(self) -> str:
        raise NotImplementedError('CSV output is not implemented')

    def _format_raw_output(self) -> str:
        raise NotImplementedError('Raw output is not implemented')



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2021, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

  Winnowing Algorithm implementation for SCANOSS.

  This module implements an adaptation of the original winnowing algorithm by S. Schleimer, D. S. Wilkerson and
  A. Aiken as described in their seminal article which can be found here:
  https://theory.stanford.edu/~aiken/publications/papers/sigmod03.pdf
"""

import hashlib
import pathlib
import platform
import re
from typing import Tuple

from binaryornot.check import is_binary
from crc32c import crc32c

from .scanossbase import ScanossBase

# Winnowing configuration. DO NOT CHANGE.
GRAM = 30
WINDOW = 64
# ASCII characters
ASCII_0 = 48
ASCII_9 = 57
ASCII_A = 65
ASCII_Z = 90
ASCII_a = 97
ASCII_z = 122
ASCII_LF = 10
ASCII_BACKSLASH = 92
MAX_CRC32 = 4294967296
MAX_LONG_LINE_CHARS = 1000
MAX_POST_SIZE = 64 * 1024  # 64k Max post size
MIN_FILE_SIZE = 256

SKIP_SNIPPET_EXT = {  # File extensions to ignore snippets for
    '.exe',
    '.zip',
    '.tar',
    '.tgz',
    '.gz',
    '.7z',
    '.rar',
    '.jar',
    '.war',
    '.ear',
    '.class',
    '.pyc',
    '.o',
    '.a',
    '.so',
    '.obj',
    '.dll',
    '.lib',
    '.out',
    '.app',
    '.bin',
    '.lst',
    '.dat',
    '.json',
    '.htm',
    '.html',
    '.xml',
    '.md',
    '.txt',
    '.doc',
    '.docx',
    '.xls',
    '.xlsx',
    '.ppt',
    '.pptx',
    '.odt',
    '.ods',
    '.odp',
    '.pages',
    '.key',
    '.numbers',
    '.pdf',
    '.min.js',
    '.mf',
    '.sum',
    '.woff',
    '.woff2',
    '.xsd',
    '.pom',
    '.whl',
}

CRC8_MAXIM_DOW_TABLE_SIZE = 0x100
CRC8_MAXIM_DOW_POLYNOMIAL = 0x8C  # 0x31 reflected
CRC8_MAXIM_DOW_INITIAL = 0x00  # 0x00 reflected
CRC8_MAXIM_DOW_FINAL = 0x00  # 0x00 reflected


class Winnowing(ScanossBase):
    """
    Winnowing Algorithm implementation for SCANOSS.

    This module implements an adaptation of the original winnowing algorithm by
    S. Schleimer, D. S. Wilkerson and A. Aiken as described in their seminal article which can be found here:
    https://theory.stanford.edu/~aiken/publications/papers/sigmod03.pdf

    The winnowing algorithm is configured using two parameters, the gram size and the window size.
    For SCANOSS the values need to be:
     - GRAM: 30
     - WINDOW: 64

    The result of performing the Winnowing algorithm is a string called WFP (Winnowing FingerPrint).
    A WFP contains optionally the name of the source component and the results of the Winnowing algorithm for each file.
    EXAMPLE output: test-component.wfp
    component=f9fc398cec3f9dd52aa76ce5b13e5f75,test-component.zip
    file=cae3ae667a54d731ca934e2867b32aaa,948,test/test-file1.c
    4=579be9fb
    5=9d9eefda,58533be6,6bb11697
    6=80188a22,f9bb9220
    10=750988e0,b6785a0d
    12=600c7ec9
    13=595544cc
    18=e3cb3b0f
    19=e8f7133d
    file=cae3ae667a54d731ca934e2867b32aaa,1843,test/test-file2.c
    2=58fb3eed
    3=f5f7f458
    4=aba6add1
    8=53762a72,0d274008,6be2454a
    10=239c7dfa
    12=0b2188c9
    15=bd9c4b10,d5c8f9fb
    16=eb7309dd,63aebec5
    19=316e10eb
    [...]

    Where component is the MD5 hash and path of the component container (could be a path to a compressed file or URL).
    file is the MD5 hash, file length and file path being fingerprinted, followed by
    a list of WFP fingerprints with their corresponding line numbers.
    """

    def __init__(  # noqa: PLR0913
        self,
        size_limit: bool = False,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
        skip_snippets: bool = False,
        post_size: int = 32,
        all_extensions: bool = False,
        obfuscate: bool = False,
        hpsm: bool = False,
        strip_hpsm_ids=None,
        strip_snippet_ids=None,
        skip_md5_ids=None,
    ):
        """
        Instantiate Winnowing class
        Parameters
        ----------
            size_limit: bool
                Limit the size of a fingerprint to 32k (post size) - Default False
        """
        super().__init__(debug, trace, quiet)
        if strip_hpsm_ids is None:
            strip_hpsm_ids = []
        if strip_snippet_ids is None:
            strip_snippet_ids = []
        if skip_md5_ids is None:
            skip_md5_ids = []
        self.size_limit = size_limit
        self.skip_snippets = skip_snippets
        self.max_post_size = post_size * 1024 if post_size > 0 else MAX_POST_SIZE
        self.all_extensions = all_extensions
        self.obfuscate = obfuscate
        self.ob_count = 1
        self.file_map = {} if obfuscate else None
        self.skip_md5_ids = skip_md5_ids
        self.strip_hpsm_ids = strip_hpsm_ids
        self.strip_snippet_ids = strip_snippet_ids
        self.hpsm = hpsm
        self.is_windows = platform.system() == 'Windows'
        if hpsm:
            self.crc8_maxim_dow_table = []
            self.crc8_generate_table()

    @staticmethod
    def __normalize(byte):
        """
        Normalise a given byte as an ASCII character
        Parameters
        ----------
        byte : int
          The byte to normalise
        """
        if byte < ASCII_0:
            return 0
        if byte > ASCII_z:
            return 0
        if byte <= ASCII_9:
            return byte
        if byte >= ASCII_a:
            return byte
        if (byte >= ASCII_A) and (byte <= ASCII_Z):
            return byte + 32
        return 0

    def __skip_snippets(self, file: str, src: str) -> bool:  # noqa: PLR0911
        """
        Determine files that are not of interest based on their content or file extension
        Parameters
        ----------
            src: str
                string to compare
        Return
        ------
            True: if file should be skipped
            False: otherwise
        """
        if self.all_extensions:
            return False
        if file:
            lower_file = file.lower()
            for ending in SKIP_SNIPPET_EXT:
                if lower_file.endswith(ending):
                    self.print_trace(f'Skipping snippets due to file ending: {file} - {ending}')
                    return True
        src_len = len(src)
        if src_len == 0 or src_len <= MIN_FILE_SIZE:  # Ignore empty or files that are too small
            self.print_trace(f'Skipping snippets as the file is too small: {file} - {src_len}')
            return True
        prefix = src[0 : (MIN_FILE_SIZE - 1)].lower().strip()
        if len(prefix) > 0 and (prefix[0] == '{' or prefix[0] == '['):  # Ignore json
            self.print_trace(f'Skipping snippets as the file appears to be JSON: {file}')
            return True
        if (
            prefix.startswith('<?xml')
            or prefix.startswith('<html')
            or prefix.startswith('<ac3d')
            or prefix.startswith('<!doc')
        ):
            self.print_trace(f'Skipping snippets as the file appears to be xml/html/binary: {file}')
            return True  # Ignore xml & html & ac3d
        index = src.index('\n') if '\n' in src else (src_len - 1)  # TODO still necessary if we have a binary check?
        if len(src[0:index]) > MAX_LONG_LINE_CHARS:  # Ignore long lines
            self.print_trace(f'Skipping snippets due to file line being too long: {file} - {MAX_LONG_LINE_CHARS}')
            return True
        return False

    def wfp_for_file(self, path: str, file: str) -> str:
        """
        Returns the WFP for a file by executing the winnowing algorithm over its contents.
        Parameters
        ----------
            path : str
                The full path of the file.
            file: str
                File name/path to record in WFP
        """
        binary_file = self.is_binary(path)
        with open(path, 'rb') as f:
            contents = f.read()
            return self.wfp_for_contents(file, binary_file, contents)

    def is_binary(self, path: str):
        """
        Check if the specified file is a potential "binary" file

        :param path: Path to the file to check
        :return: True if binary, False otherwise
        """
        if path:
            binary_path = is_binary(path)
            if binary_path:
                self.print_trace(f'Detected binary file: {path}')
            return binary_path
        return False

    def __strip_hpsm(self, file: str, hpsm: str) -> str:
        """
        Strip off request HPSM IDs if requested

        :param file: name of the fingerprinted file
        :param hpsm:  HPSM string
        :return: modified HPSM (if necessary)
        """
        hpsm_len = len(hpsm)
        if self.strip_hpsm_ids and hpsm_len > 1:
            # Check for HPSM ID strings to remove. The size of the sequence must be conserved.
            for hpsm_id in self.strip_hpsm_ids:
                hpsm_id_index = hpsm.find(hpsm_id)
                hpsm_id_len = len(hpsm_id)
                # If the position is odd, we need to overwrite one byte before.
                if hpsm_id_index % 2 == 1:
                    hpsm_id_index = hpsm_id_index - 1
                    # If the size of the sequence is even, we need to overwrite one byte after.
                    if hpsm_id_len % 2 == 0:
                        hpsm_id_len = hpsm_id_len + 1
                    hpsm_id_len = hpsm_id_len + 1
                # If the position is even and the size is odd, we need to overwrite one byte after.
                elif hpsm_id_len % 2 == 1:
                    hpsm_id_len = hpsm_id_len + 1

                to_remove = hpsm[hpsm_id_index : hpsm_id_index + hpsm_id_len]
                self.print_debug(f'HPSM ID {to_remove} to replace')
                # Calculate the XOR of each byte to produce the correct ignore sequence.
                replacement = ''.join(
                    [format(int(to_remove[i : i + 2], 16) ^ 0xFF, '02x') for i in range(0, len(to_remove), 2)]
                )

                self.print_debug(f'HPSM ID replacement {replacement}')
                # Overwrite HPSM bytes to be removed.
                hpsm = hpsm.replace(to_remove, replacement)
            if hpsm_len != len(hpsm):
                self.print_stderr(f'wrong HPSM values from {file}')
        return hpsm

    def __strip_snippets(self, file: str, wfp: str) -> str:
        """
        Strip snippet IDs from the WFP

        :param file: name of fingerprinted file
        :param wfp: WFP to clean
        :return: updated WFP
        """
        wfp_len = len(wfp)
        for snippet_id in self.strip_snippet_ids:  # Remove exact snippet strings
            wfp = wfp.replace(snippet_id, '')
        if wfp_len > len(wfp):
            wfp = re.sub(r'(,)\1+', ',', wfp)  # Remove multiple 'empty comma' blocks
            wfp = wfp.replace(',\n', '\n')  # Remove trailing comma
            wfp = wfp.replace('=,', '=')  # Remove leading comma
            wfp = re.sub(r'\d+=\s+', '', wfp)  # Cleanup empty lines
            self.print_debug(f'Stripped snippet ids from {file}')
        return wfp

    def __detect_line_endings(self, contents: bytes) -> Tuple[bool, bool, bool]:
        """Detect the types of line endings present in file contents.

        Args:
            contents: File contents as bytes.

        Returns:
            Tuple of (has_crlf, has_lf_only, has_cr_only, has_mixed) indicating which line ending types are present.
        """
        has_crlf = b'\r\n' in contents
        # For LF detection, we need to find LF that's not part of CRLF
        content_without_crlf = contents.replace(b'\r\n', b'')
        has_standalone_lf = b'\n' in content_without_crlf
        # For CR detection, we need to find CR that's not part of CRLF
        has_standalone_cr = b'\r' in content_without_crlf

        return has_crlf, has_standalone_lf, has_standalone_cr

    def __calculate_opposite_line_ending_hash(self, contents: bytes):
        """Calculate hash for contents with opposite line endings.

        If the file is primarily Unix (LF), calculates Windows (CRLF) hash.
        If the file is primarily Windows (CRLF), calculates Unix (LF) hash.

        Args:
            contents: File contents as bytes.

        Returns:
            Hash with opposite line endings as hex string, or None if no line endings detected.
        """
        has_crlf, has_standalone_lf, has_standalone_cr = self.__detect_line_endings(contents)

        if not has_crlf and not has_standalone_lf and not has_standalone_cr:
            return None

        # Normalize all line endings to LF first
        normalized = contents.replace(b'\r\n', b'\n').replace(b'\r', b'\n')

        # Determine the dominant line ending type
        if has_crlf and not has_standalone_lf and not has_standalone_cr:
            # File is Windows (CRLF) - produce Unix (LF) hash
            opposite_contents = normalized
        else:
            # File is Unix (LF/CR) or mixed - produce Windows (CRLF) hash
            opposite_contents = normalized.replace(b'\n', b'\r\n')

        return hashlib.md5(opposite_contents).hexdigest()

    def wfp_for_contents(self, file: str, bin_file: bool, contents: bytes) -> str:  # noqa: PLR0912, PLR0915
        """
        Generate a Winnowing fingerprint (WFP) for the given file contents
        Parameters
        ----------
            :param file: file to fingerprint
            :param bin_file: binary file or not
            :param contents: file contents
        Return
        ------
            WFP string
        """
        file_md5 = hashlib.md5(contents).hexdigest()
        if self.skip_md5_ids and file_md5 in self.skip_md5_ids:
            self.print_debug(f'Skipping MD5 file name for {file_md5}: {file}')
            return ''
        # Print file line
        content_length = len(contents)
        original_filename = file

        if self.is_windows:
            original_filename = file.replace('\\', '/')
        wfp_filename = repr(original_filename).strip("'")  # return a utf-8 compatible version of the filename
        if self.obfuscate:  # hide the real size of the file and its name, but keep the suffix
            wfp_filename = f'{self.ob_count}{pathlib.Path(original_filename).suffix}'
            self.ob_count = self.ob_count + 1
            self.file_map[wfp_filename] = original_filename  # Save the file name map for later (reverse lookup)

        wfp = 'file={0},{1},{2}\n'.format(file_md5, content_length, wfp_filename)

        # Add opposite line ending hash based on line ending analysis
        if not bin_file:
            opposite_hash = self.__calculate_opposite_line_ending_hash(contents)
            if opposite_hash is not None:
                wfp += f'fh2={opposite_hash}\n'

        # We don't process snippets for binaries, or other uninteresting files, or if we're requested to skip
        if bin_file or self.skip_snippets or self.__skip_snippets(file, contents.decode('utf-8', 'ignore')):
            return wfp
        # Add HPSM
        if self.hpsm:
            hpsm = self.__strip_hpsm(file, self.calc_hpsm(contents))
            if len(hpsm) > 0:
                wfp += f'hpsm={hpsm}\n'
        # Initialize variables
        gram = ''
        window = []
        line = 1
        last_hash = MAX_CRC32
        last_line = 0
        output = ''
        # Otherwise, recurse src_content and calculate Winnowing hashes
        for byte in contents:
            if byte == ASCII_LF:
                line += 1
                normalized = 0
            else:
                normalized = self.__normalize(byte)
            # Is it a useful byte?
            if normalized:
                gram += chr(normalized)  # Add byte to gram
                # Do we have a full gram?
                if len(gram) >= GRAM:
                    gram_crc32 = crc32c(gram.encode('ascii'))
                    window.append(gram_crc32)
                    # Do we have a full window?
                    if len(window) >= WINDOW:
                        # Select minimum hash for the current window
                        min_hash = min(window)
                        # Is the minimum hash a new one?
                        if min_hash != last_hash:
                            # Hashing the hash will result in a better balanced resulting data set
                            # as it will counter the winnowing effect which selects the "minimum"
                            # hash in each window
                            crc = crc32c(min_hash.to_bytes(4, byteorder='little'))
                            crc_hex = '{:08x}'.format(crc)
                            if last_line != line:
                                if output != '':
                                    if (
                                        self.size_limit
                                        and (len(wfp.encode('utf-8')) + len(output.encode('utf-8')))
                                        > self.max_post_size
                                    ):
                                        self.print_debug(f'Truncating WFP ({self.max_post_size} limit) for: {file}')
                                        output = ''
                                        break  # Stop collecting snippets as it's over 64k
                                    wfp += output + '\n'
                                output = '%d=%s' % (line, crc_hex)
                            else:
                                output += ',' + crc_hex

                            last_line = line
                            last_hash = min_hash
                        # Shift window
                        window.pop(0)
                    # Shift gram
                    gram = gram[1:]
        if output != '':
            if not self.size_limit or (len(wfp.encode('utf-8')) + len(output.encode('utf-8'))) < self.max_post_size:
                wfp += output + '\n'
            else:
                self.print_debug(f'Warning: skipping output in WFP for {file} - "{output}"')

        if wfp is None or wfp == '':
            self.print_stderr(f'Warning: No WFP content data for {file}')
        elif self.strip_snippet_ids:
            wfp = self.__strip_snippets(file, wfp)

        return wfp

    def calc_hpsm(self, content):
        """
        Calculate the HPSM data for this content

        :param content: content bytes to calculate
        :return: HPSM encoded data
        """
        list_normalized = []  # Array of numbers
        crc_lines = []  # Array of numbers that represent the crc8_maxim for each line of the file
        last_line = 0
        for i, byte in enumerate(content):
            c = byte
            if c == ASCII_LF:  # When there is a new line
                if list_normalized:
                    crc_lines.append(self.crc8_buffer(list_normalized))
                    list_normalized = []
                elif last_line + 1 == i:
                    crc_lines.append(0xFF)
                elif i - last_line > 1:
                    crc_lines.append(0x00)
                last_line = i
            else:
                c_normalized = self.__normalize(c)
                if c_normalized != 0:
                    list_normalized.append(c_normalized)
        hpsm = ''.join('{:02x}'.format(x) for x in crc_lines)
        return hpsm

    def crc8_generate_table(self):
        """
        Generate the CRC8 maxim dow table

        :return: nothing
        """
        if not self.crc8_maxim_dow_table or len(self.crc8_maxim_dow_table) == 0:
            for i in range(CRC8_MAXIM_DOW_TABLE_SIZE):
                self.crc8_maxim_dow_table.append(self.crc8_byte_checksum(0, i))

    @staticmethod
    def crc8_byte_checksum(crc: int, byte):
        """
        Calculate the CRC8 checksum for the given byte

        :param crc:
        :param byte:
        :return: CRC for the byte
        """
        crc ^= byte
        for count in range(8):
            is_set = crc & 0x01
            crc >>= 1
            if is_set:
                crc ^= CRC8_MAXIM_DOW_POLYNOMIAL
        return crc

    def crc8_byte(self, crc: int, byte):
        """
        Calculate the CRC8 for the given CRC & Byte

        :param crc:
        :param byte:
        :return:
        """
        index = byte ^ crc
        return self.crc8_maxim_dow_table[index] ^ (crc >> 8)

    def crc8_buffer(self, buffer):
        """
        Calculate the CRC for the given buffer list

        :param buffer:
        :return:
        """
        crc = CRC8_MAXIM_DOW_INITIAL
        for index in range(len(buffer)):
            crc = self.crc8_byte(crc, buffer[index])
        crc ^= CRC8_MAXIM_DOW_FINAL  # Bitwise OR (XOR) of crc in Maxim Dow Final
        return crc


#
# End of Winnowing Class
#



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2025, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import json
import threading
import time
from typing import Dict, Optional

from progress.spinner import Spinner

from scanoss.constants import DEFAULT_HFH_RANK_THRESHOLD
from scanoss.cyclonedx import CycloneDx
from scanoss.file_filters import FileFilters
from scanoss.scanners.folder_hasher import FolderHasher
from scanoss.scanners.scanner_config import ScannerConfig
from scanoss.scanoss_settings import ScanossSettings
from scanoss.scanossbase import ScanossBase
from scanoss.scanossgrpc import ScanossGrpc
from scanoss.utils.abstract_presenter import AbstractPresenter


class ScannerHFH:
    """
    Folder Hashing Scanner.

    This scanner processes a directory, computes CRC64 hashes for the files,
    and calculates simhash values based on file names and content to detect folder-level similarities.
    """

    def __init__(
        self,
        scan_dir: str,
        config: ScannerConfig,
        client: Optional[ScanossGrpc] = None,
        scanoss_settings: Optional[ScanossSettings] = None,
        rank_threshold: int = DEFAULT_HFH_RANK_THRESHOLD,
    ):
        """
        Initialize the ScannerHFH.

        Args:
            scan_dir (str): The directory to be scanned.
            config (ScannerConfig): Configuration parameters for the scanner.
            client (ScanossGrpc): gRPC client for communicating with the scanning service.
            scanoss_settings (Optional[ScanossSettings]): Optional settings for Scanoss.
            rank_threshold (int): Get results with rank below this threshold (default: 5).
        """
        self.base = ScanossBase(
            debug=config.debug,
            trace=config.trace,
            quiet=config.quiet,
        )
        self.presenter = ScannerHFHPresenter(
            self,
            debug=config.debug,
            trace=config.trace,
            quiet=config.quiet,
        )
        self.file_filters = FileFilters(
            debug=config.debug,
            trace=config.trace,
            quiet=config.quiet,
            scanoss_settings=scanoss_settings,
        )
        self.folder_hasher = FolderHasher(
            scan_dir=scan_dir,
            config=config,
            scanoss_settings=scanoss_settings,
        )

        self.scan_dir = scan_dir
        self.client = client
        self.scan_results = None
        self.rank_threshold = rank_threshold

    def scan(self) -> Optional[Dict]:
        """
        Scan the provided directory using the folder hashing algorithm.

        Returns:
            Optional[Dict]: The folder hash response from the gRPC client, or None if an error occurs.
        """
        hfh_request = {
            'root': self.folder_hasher.hash_directory(self.scan_dir),
            'rank_threshold': self.rank_threshold,
        }

        spinner = Spinner('Scanning folder...')
        stop_spinner = False

        def spin():
            while not stop_spinner:
                spinner.next()
                time.sleep(0.1)

        spinner_thread = threading.Thread(target=spin)
        spinner_thread.start()

        try:
            response = self.client.folder_hash_scan(hfh_request)
            if response:
                self.scan_results = response
        finally:
            stop_spinner = True
            spinner_thread.join()
            spinner.finish()

        return self.scan_results

    def present(self, output_format: str = None, output_file: str = None):
        """Present the results in the selected format"""
        self.presenter.present(output_format=output_format, output_file=output_file)


class ScannerHFHPresenter(AbstractPresenter):
    """
    ScannerHFH presenter class
    Handles the presentation of the folder hashing scan results
    """

    def __init__(self, scanner: ScannerHFH, **kwargs):
        super().__init__(**kwargs)
        self.scanner = scanner

    def _format_json_output(self) -> str:
        """
        Format the scan output data into a JSON object

        Returns:
            str: The formatted JSON string
        """
        return json.dumps(self.scanner.scan_results, indent=2)

    def _format_plain_output(self) -> str:
        """
        Format the scan output data into a plain text string
        """
        return (
            json.dumps(self.scanner.scan_results, indent=2)
            if isinstance(self.scanner.scan_results, dict)
            else str(self.scanner.scan_results)
        )

    def _format_cyclonedx_output(self) -> str:  # noqa: PLR0911
        if not self.scanner.scan_results:
            return ''
        try:
            if 'results' not in self.scanner.scan_results or not self.scanner.scan_results['results']:
                self.base.print_stderr('ERROR: No scan results found')
                return ''

            first_result = self.scanner.scan_results['results'][0]

            best_match_components = [c for c in first_result.get('components', []) if c.get('order') == 1]
            if not best_match_components:
                self.base.print_stderr('ERROR: No best match component found')
                return ''

            best_match_component = best_match_components[0]
            if not best_match_component.get('versions'):
                self.base.print_stderr('ERROR: No versions found for best match component')
                return ''

            best_match_version = best_match_component['versions'][0]
            purl = best_match_component['purl']

            get_dependencies_json_request = {
                'files': [
                    {
                        'file': f'{best_match_component["name"]}:{best_match_version["version"]}',
                        'purls': [{'purl': purl, 'requirement': best_match_version['version']}],
                    }
                ]
            }
            
            get_vulnerabilities_json_request = {
                'purls': [{'purl': purl, 'requirement': best_match_version['version']}],
            }

            decorated_scan_results = self.scanner.client.get_dependencies(get_dependencies_json_request)
            vulnerabilities = self.scanner.client.get_vulnerabilities_json(get_vulnerabilities_json_request)

            cdx = CycloneDx(self.base.debug)
            scan_results = {}
            for f in decorated_scan_results['files']:
                scan_results[f['file']] = [f]
            success, cdx_output = cdx.produce_from_json(scan_results)
            if not success:
                error_msg = 'ERROR: Failed to produce CycloneDX output'
                self.base.print_stderr(error_msg)
                return None
            
            if vulnerabilities:
                cdx_output = cdx.append_vulnerabilities(cdx_output, vulnerabilities, purl)
                
            return json.dumps(cdx_output, indent=2)
        except Exception as e:
            self.base.print_stderr(f'ERROR: Failed to get license information: {e}')
            return None

    def _format_spdxlite_output(self) -> str:
        raise NotImplementedError('SPDXlite output is not implemented')

    def _format_csv_output(self) -> str:
        raise NotImplementedError('CSV output is not implemented')

    def _format_raw_output(self) -> str:
        raise NotImplementedError('Raw output is not implemented')
